# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import datetime
import io
import math
import os
import logging
from typing import Mapping, Optional, Sequence, Dict, Tuple, List

from pyomo.common.collections import ComponentMap
from pyomo.common.config import ConfigValue
from pyomo.common.dependencies import attempt_import
from pyomo.common.enums import ObjectiveSense
from pyomo.common.errors import ApplicationError
from pyomo.common.shutdown import python_is_shutting_down
from pyomo.common.tee import capture_output, TeeStream
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.base import VarData, ConstraintData

from pyomo.contrib.solver.common.base import SolverBase, Availability
from pyomo.contrib.solver.common.config import BranchAndBoundConfig
from pyomo.contrib.solver.common.util import (
    NoFeasibleSolutionError,
    NoOptimalSolutionError,
    NoDualsError,
    NoReducedCostsError,
    NoSolutionError,
)
from pyomo.contrib.solver.common.results import (
    Results,
    SolutionStatus,
    TerminationCondition,
)
from pyomo.contrib.solver.common.solution_loader import SolutionLoaderBase
import time

logger = logging.getLogger(__name__)

gurobipy, gurobipy_available = attempt_import('gurobipy')


class GurobiConfig(BranchAndBoundConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        BranchAndBoundConfig.__init__(
            self,
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )
        self.warmstart_discrete_vars: bool = self.declare(
            'warmstart_discrete_vars',
            ConfigValue(
                default=False,
                domain=bool,
                description="If True, the current values of the integer variables "
                "will be passed to Gurobi.",
            ),
        )


class GurobiDirectSolutionLoaderBase(SolutionLoaderBase):
    def __init__(self, solver_model) -> None:
        super().__init__()
        self._solver_model = solver_model
        GurobiDirectBase._register_env_client()

    def _get_var_lists(self):
        """
        Should return a list of pyomo vars and a list of gurobipy vars
        """
        raise NotImplementedError('should be implemented by derived classes')

    def _get_var_map(self):
        raise NotImplementedError('should be implemented by derived classes')

    def _get_con_map(self):
        raise NotImplementedError('should be implemented by derived classes')

    def __del__(self):
        # Release the gurobi license if this is the last reference to
        # the environment (either through a results object or solver
        # interface)
        GurobiDirectBase._release_env_client()

    def _get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None, solution_id=0
    ) -> Tuple[List[VarData], List[float]]:
        if self._solver_model.SolCount == 0:
            raise NoSolutionError()
        if vars_to_load is None:
            pvars, gvars = self._get_var_lists()
        else:
            pvars = vars_to_load
            gvars = list(map(self._get_var_map().__getitem__, vars_to_load))
        if solution_id:
            if (
                self._solver_model.getAttr('NumIntVars') == 0
                and self._solver_model.getAttr('NumBinVars') == 0
            ):
                raise ValueError(
                    'Cannot obtain suboptimal solutions for a continuous model'
                )
            original_solution_number = self._solver_model.getParamInfo(
                'SolutionNumber'
            )[2]
            self._solver_model.setParam('SolutionNumber', solution_id)
            grbFcn = "Xn"
        else:
            grbFcn = "X"
        try:
            vals = self._solver_model.getAttr(grbFcn, gvars)
        finally:
            if solution_id:
                self._solver_model.setParam('SolutionNumber', original_solution_number)
        return pvars, vals

    def load_vars(
        self, vars_to_load: Optional[Sequence[VarData]] = None, solution_id=0
    ) -> None:
        pvars, vals = self._get_primals(
            vars_to_load=vars_to_load, solution_id=solution_id
        )
        for pv, val in zip(pvars, vals):
            pv.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None, solution_id=0
    ) -> Mapping[VarData, float]:
        pvars, vals = self._get_primals(
            vars_to_load=vars_to_load, solution_id=solution_id
        )
        res = ComponentMap(zip(pvars, vals))
        return res

    def _get_rc_all_vars(self):
        pvars, gvars = self._get_var_lists()
        vals = self._solver_model.getAttr("Rc", gvars)
        return ComponentMap((i, val) for i, val in zip(pvars, vals))

    def _get_rc_subset_vars(self, vars_to_load):
        var_map = self._get_var_map()
        gvars = [var_map[i] for i in vars_to_load]
        vals = self._solver_model.getAttr("Rc", gvars)
        return ComponentMap(zip(vars_to_load, vals))

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._solver_model.Status != gurobipy.GRB.OPTIMAL:
            raise NoReducedCostsError()
        if self._solver_model.IsMIP:
            # this will also return True for continuous, nonconvex models
            raise NoReducedCostsError()
        if vars_to_load is None:
            res = self._get_rc_all_vars()
        else:
            res = self._get_rc_subset_vars(vars_to_load=vars_to_load)
        return res

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        if self._solver_model.Status != gurobipy.GRB.OPTIMAL:
            raise NoDualsError()
        if self._solver_model.IsMIP:
            # this will also return True for continuous, nonconvex models
            raise NoDualsError()

        qcons = set(self._solver_model.getQConstrs())
        con_map = self._get_con_map()
        if cons_to_load is None:
            cons_to_load = con_map.keys()

        duals = {}
        for c in cons_to_load:
            gurobi_con = con_map[c]
            if type(gurobi_con) is tuple:
                # only linear range constraints are supported
                gc1, gc2 = gurobi_con
                d1 = gc1.Pi
                d2 = gc2.Pi
                if abs(d1) > abs(d2):
                    duals[c] = d1
                else:
                    duals[c] = d2
            else:
                if gurobi_con in qcons:
                    duals[c] = gurobi_con.QCPi
                else:
                    duals[c] = gurobi_con.Pi

        return duals


class GurobiDirectBase(SolverBase):

    _num_gurobipy_env_clients = 0
    _gurobipy_env = None
    _available = None
    _gurobipy_available = gurobipy_available
    _tc_map = None
    _minimum_version = (0, 0, 0)

    CONFIG = GurobiConfig()

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self._register_env_client()
        self._callback = None

    def __del__(self):
        if not python_is_shutting_down():
            self._release_env_client()

    def available(self):
        if self._available is None:
            # this triggers the deferred import, and for the persistent
            # interface, may update the _available flag
            #
            # Note that we set the _available flag on the *most derived
            # class* and not on the instance, or on the base class.  That
            # allows different derived interfaces to have different
            # availability (e.g., persistent has a minimum version
            # requirement that the direct interface doesn't - is that true?)
            if not self._gurobipy_available:
                if self._available is None:
                    self.__class__._available = Availability.NotFound
            else:
                self.__class__._available = self._check_license()
                if self.version() < self._minimum_version:
                    self.__class__._available = Availability.BadVersion
        return self._available

    @staticmethod
    def release_license():
        if GurobiDirectBase._gurobipy_env is None:
            return
        if GurobiDirectBase._num_gurobipy_env_clients:
            logger.warning(
                "Call to GurobiDirectBase.release_license() with %s remaining "
                "environment clients." % (GurobiDirectBase._num_gurobipy_env_clients,)
            )
        GurobiDirectBase._gurobipy_env.close()
        GurobiDirectBase._gurobipy_env = None

    @staticmethod
    def env():
        if GurobiDirectBase._gurobipy_env is None:
            with capture_output(capture_fd=True):
                GurobiDirectBase._gurobipy_env = gurobipy.Env()
        return GurobiDirectBase._gurobipy_env

    @staticmethod
    def _register_env_client():
        GurobiDirectBase._num_gurobipy_env_clients += 1

    @staticmethod
    def _release_env_client():
        GurobiDirectBase._num_gurobipy_env_clients -= 1
        if GurobiDirectBase._num_gurobipy_env_clients <= 0:
            # Note that _num_gurobipy_env_clients should never be <0,
            # but if it is, release_license will issue a warning (that
            # we want to know about)
            GurobiDirectBase.release_license()

    def _check_license(self):
        try:
            model = gurobipy.Model(env=self.env())
        except gurobipy.GurobiError:
            return Availability.BadLicense

        model.setParam('OutputFlag', 0)
        try:
            model.addVars(range(2001))
            model.optimize()
            return Availability.FullLicense
        except gurobipy.GurobiError:
            return Availability.LimitedLicense
        finally:
            model.dispose()

    def version(self):
        if not gurobipy_available:
            return None
        version = (
            gurobipy.GRB.VERSION_MAJOR,
            gurobipy.GRB.VERSION_MINOR,
            gurobipy.GRB.VERSION_TECHNICAL,
        )
        return version

    def _create_solver_model(self, pyomo_model, config):
        # should return gurobi_model, solution_loader, has_objective
        raise NotImplementedError('should be implemented by derived classes')

    def _pyomo_gurobi_var_iter(self):
        # generator of tuples (pyomo_var, gurobi_var)
        raise NotImplementedError('should be implemented by derived classes')

    def _mipstart(self):
        for pyomo_var, gurobi_var in self._pyomo_gurobi_var_iter():
            if pyomo_var.is_integer() and pyomo_var.value is not None:
                gurobi_var.setAttr('Start', pyomo_var.value)

    def solve(self, model, **kwds) -> Results:
        start_timestamp = datetime.datetime.now(datetime.timezone.utc)
        tick = time.perf_counter()
        orig_cwd = os.getcwd()
        try:
            config = self.config(value=kwds, preserve_implicit=True)

            if config.timer is None:
                config.timer = HierarchicalTimer()
            timer = config.timer

            StaleFlagManager.mark_all_as_stale()
            ostreams = [io.StringIO()] + config.tee

            if config.working_dir:
                os.chdir(config.working_dir)
            with capture_output(TeeStream(*ostreams), capture_fd=False):
                gurobi_model, solution_loader, has_obj = self._create_solver_model(
                    model, config
                )
                options = config.solver_options

                gurobi_model.setParam('LogToConsole', 1)

                if config.threads is not None:
                    gurobi_model.setParam('Threads', config.threads)
                if config.time_limit is not None:
                    gurobi_model.setParam('TimeLimit', config.time_limit)
                if config.rel_gap is not None:
                    gurobi_model.setParam('MIPGap', config.rel_gap)
                if config.abs_gap is not None:
                    gurobi_model.setParam('MIPGapAbs', config.abs_gap)

                if config.warmstart_discrete_vars:
                    self._mipstart()

                for key, option in options.items():
                    gurobi_model.setParam(key, option)

                timer.start('optimize')
                gurobi_model.optimize(self._callback)
                timer.stop('optimize')

            res = self._populate_results(
                grb_model=gurobi_model,
                solution_loader=solution_loader,
                has_obj=has_obj,
                config=config,
            )
        finally:
            os.chdir(orig_cwd)

        res.solver_log = ostreams[0].getvalue()
        tock = time.perf_counter()
        res.timing_info.start_timestamp = start_timestamp
        res.timing_info.wall_time = tock - tick
        res.timing_info.timer = timer
        return res

    def _get_tc_map(self):
        if GurobiDirectBase._tc_map is None:
            grb = gurobipy.GRB
            tc = TerminationCondition
            GurobiDirectBase._tc_map = {
                grb.LOADED: tc.unknown,  # problem is loaded, but no solution
                grb.OPTIMAL: tc.convergenceCriteriaSatisfied,
                grb.INFEASIBLE: tc.provenInfeasible,
                grb.INF_OR_UNBD: tc.infeasibleOrUnbounded,
                grb.UNBOUNDED: tc.unbounded,
                grb.CUTOFF: tc.objectiveLimit,
                grb.ITERATION_LIMIT: tc.iterationLimit,
                grb.NODE_LIMIT: tc.iterationLimit,
                grb.TIME_LIMIT: tc.maxTimeLimit,
                grb.SOLUTION_LIMIT: tc.unknown,
                grb.INTERRUPTED: tc.interrupted,
                grb.NUMERIC: tc.unknown,
                grb.SUBOPTIMAL: tc.unknown,
                grb.USER_OBJ_LIMIT: tc.objectiveLimit,
            }
        return GurobiDirectBase._tc_map

    def _populate_results(self, grb_model, solution_loader, has_obj, config):
        status = grb_model.Status

        results = Results()
        results.solution_loader = solution_loader
        results.timing_info.gurobi_time = grb_model.Runtime

        if grb_model.SolCount > 0:
            if status == gurobipy.GRB.OPTIMAL:
                results.solution_status = SolutionStatus.optimal
            else:
                results.solution_status = SolutionStatus.feasible
        else:
            results.solution_status = SolutionStatus.noSolution

        results.termination_condition = self._get_tc_map().get(
            status, TerminationCondition.unknown
        )

        if (
            results.termination_condition
            != TerminationCondition.convergenceCriteriaSatisfied
            and config.raise_exception_on_nonoptimal_result
        ):
            raise NoOptimalSolutionError()

        if has_obj:
            try:
                if math.isfinite(grb_model.ObjVal):
                    results.incumbent_objective = grb_model.ObjVal
                else:
                    results.incumbent_objective = None
            except (gurobipy.GurobiError, AttributeError):
                results.incumbent_objective = None
            try:
                results.objective_bound = grb_model.ObjBound
            except (gurobipy.GurobiError, AttributeError):
                if grb_model.ModelSense == ObjectiveSense.minimize:
                    results.objective_bound = -math.inf
                else:
                    results.objective_bound = math.inf
        else:
            results.incumbent_objective = None
            results.objective_bound = None

        results.extra_info.IterCount = grb_model.getAttr('IterCount')
        results.extra_info.BarIterCount = grb_model.getAttr('BarIterCount')
        results.extra_info.NodeCount = grb_model.getAttr('NodeCount')

        config.timer.start('load solution')
        if config.load_solutions:
            if grb_model.SolCount > 0:
                results.solution_loader.load_vars()
            else:
                raise NoFeasibleSolutionError()
        config.timer.stop('load solution')

        results.solver_config = config
        results.solver_name = self.name
        results.solver_version = self.version()

        return results



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as np, attempt_import

block_vector = attempt_import('pyomo.contrib.pynumero.sparse.block_vector')[0]


def norm(x, ord=None):
    f = np.linalg.norm
    if isinstance(x, np.ndarray):
        return f(x, ord=ord)
    elif isinstance(x, block_vector.BlockVector):
        flat_x = x.flatten()
        return f(flat_x, ord=ord)
    else:
        raise NotImplementedError()


def allclose(x1, x2, rtol, atol):
    # this needs to be implemented for parallel
    x1_flat = x1.flatten()
    x2_flat = x2.flatten()
    return np.allclose(x1_flat, x2_flat, rtol=rtol, atol=atol)


def concatenate(arrays):
    return np.concatenate([a.flatten() for a in arrays])


def where(*args):
    condition = args[0]

    if len(args) == 2:
        raise ValueError('either both or neither of x and y should be given')
    if len(args) > 3:
        raise TypeError(
            'where() takes at most 3 arguments ({} given)'.format(len(args))
        )

    n_args = len(args)
    if isinstance(condition, block_vector.BlockVector):
        if n_args == 1:
            assert (
                not condition.has_none
            ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
            res = block_vector.BlockVector(condition.nblocks)
            for i in range(condition.nblocks):
                _args = [condition.get_block(i)]
                res.set_block(i, where(*_args)[0])
            return (res,)
        else:
            x = args[1]
            y = args[2]
            if isinstance(x, block_vector.BlockVector) and isinstance(
                y, block_vector.BlockVector
            ):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    not x.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    not y.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    condition.nblocks == x.nblocks
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.nblocks == y.nblocks
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                for i in range(condition.nblocks):
                    _args = [condition.get_block(i), x.get_block(i), y.get_block(i)]
                    res.set_block(i, where(*_args))
                return res
            elif isinstance(x, np.ndarray) and isinstance(y, block_vector.BlockVector):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    not y.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    condition.nblocks == y.nblocks
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.size == condition.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.size == y.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [
                        condition.get_block(i),
                        x[accum : accum + nelements],
                        y.get_block(i),
                    ]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif isinstance(x, block_vector.BlockVector) and isinstance(y, np.ndarray):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    not x.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    condition.nblocks == x.nblocks
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.size == condition.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.size == y.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [
                        condition.get_block(i),
                        x.get_block(i),
                        y[accum : accum + nelements],
                    ]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif np.isscalar(x) and isinstance(y, block_vector.BlockVector):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    not y.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    condition.nblocks == y.nblocks
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    condition.size == y.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [condition.get_block(i), x, y.get_block(i)]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif isinstance(x, block_vector.BlockVector) and np.isscalar(y):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    not x.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    condition.nblocks == x.nblocks
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.size == condition.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [condition.get_block(i), x.get_block(i), y]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    x.size == condition.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                assert (
                    x.size == y.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [
                        condition.get_block(i),
                        x[accum : accum + nelements],
                        y[accum : accum + nelements],
                    ]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif isinstance(x, np.ndarray) and np.isscalar(y):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    x.size == condition.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [condition.get_block(i), x[accum : accum + nelements], y]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif np.isscalar(x) and isinstance(y, np.ndarray):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                assert (
                    condition.size == y.size
                ), 'Operation on BlockVectors need the same number of blocks on each operand'
                res = block_vector.BlockVector(condition.nblocks)
                accum = 0
                for i in range(condition.nblocks):
                    nelements = condition._brow_lengths[i]
                    _args = [condition.get_block(i), x, y[accum : accum + nelements]]
                    res.set_block(i, where(*_args))
                    accum += nelements
                return res

            elif np.isscalar(x) and np.isscalar(y):
                assert (
                    not condition.has_none
                ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
                res = block_vector.BlockVector(condition.nblocks)
                for i in range(condition.nblocks):
                    _args = [condition.get_block(i), x, y]
                    res.set_block(i, where(*_args))
                return res

            else:
                raise NotImplementedError()
    else:
        if n_args == 1:
            return np.where(*args)
        else:
            x = args[1]
            y = args[2]
            if isinstance(x, block_vector.BlockVector):
                # ToDo: add logger to give warning here
                x = x.flatten()
            if isinstance(y, block_vector.BlockVector):
                # ToDo: add logger to give warning here
                y = y.flatten()
            _args = [condition, x, y]
            return np.where(*_args)


def isin(element, test_elements, assume_unique=False, invert=False):
    if isinstance(element, block_vector.BlockVector) and isinstance(
        test_elements, block_vector.BlockVector
    ):
        assert (
            not element.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
        assert (
            not test_elements.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
        assert (
            element.nblocks == test_elements.nblocks
        ), 'Operation on BlockVectors need the same number of blocks on each operand'
        res = block_vector.BlockVector(element.nblocks)
        for i in range(element.nblocks):
            res.set_block(
                i,
                isin(
                    element.get_block(i),
                    test_elements.get_block(i),
                    assume_unique=assume_unique,
                    invert=invert,
                ),
            )
        return res

    elif isinstance(element, block_vector.BlockVector) and isinstance(
        test_elements, np.ndarray
    ):
        assert (
            not element.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
        res = block_vector.BlockVector(element.nblocks)
        for i in range(element.nblocks):
            res.set_block(
                i,
                isin(
                    element.get_block(i),
                    test_elements,
                    assume_unique=assume_unique,
                    invert=invert,
                ),
            )
        return res

    elif isinstance(element, np.ndarray) and isinstance(test_elements, np.ndarray):
        return np.isin(
            element, test_elements, assume_unique=assume_unique, invert=invert
        )

    else:
        raise NotImplementedError()


def intersect1d(ar1, ar2, assume_unique=False, return_indices=False):
    if return_indices:
        raise NotImplementedError()

    if isinstance(ar1, tuple) and len(ar1) == 1:
        x = ar1[0]
    elif isinstance(ar1, np.ndarray) or isinstance(ar1, block_vector.BlockVector):
        x = ar1
    else:
        raise RuntimeError(
            'ar1 type not recognized. Needs to be np.ndarray or BlockVector'
        )

    if isinstance(ar2, tuple) and len(ar2) == 1:
        y = ar2[0]
    elif isinstance(ar2, np.ndarray) or isinstance(ar1, block_vector.BlockVector):
        y = ar2
    else:
        raise RuntimeError(
            'ar2 type not recognized. Needs to be np.ndarray or BlockVector'
        )

    if isinstance(x, block_vector.BlockVector) and isinstance(
        y, block_vector.BlockVector
    ):
        assert x.nblocks == y.nblocks, "Number of blocks does not match"
        assert (
            not x.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
        assert (
            not y.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'

        res = block_vector.BlockVector(x.nblocks)
        for i in range(x.nblocks):
            res.set_block(
                i,
                intersect1d(
                    x.get_block(i), y.get_block(i), assume_unique=assume_unique
                ),
            )
        return res
    elif isinstance(x, block_vector.BlockVector) and isinstance(y, np.ndarray):
        assert (
            not x.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'

        res = block_vector.BlockVector(x.nblocks)
        for i in range(x.nblocks):
            res.set_block(
                i, np.intersect1d(x.get_block(i), y, assume_unique=assume_unique)
            )
        return res
    elif isinstance(x, np.ndarray) and isinstance(y, block_vector.BlockVector):
        assert (
            not y.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'

        res = block_vector.BlockVector(y.nblocks)
        for i in range(y.nblocks):
            res.set_block(
                i, np.intersect1d(x, y.get_block(i), assume_unique=assume_unique)
            )
        return res
    else:
        return np.intersect1d(x, y, assume_unique=assume_unique)


def setdiff1d(ar1, ar2, assume_unique=False):
    if isinstance(ar1, tuple) and len(ar1) == 1:
        x = ar1[0]
    elif isinstance(ar1, np.ndarray) or isinstance(ar1, block_vector.BlockVector):
        x = ar1
    else:
        raise RuntimeError(
            'ar1 type not recognized. Needs to be np.ndarray or BlockVector'
        )

    if isinstance(ar2, tuple) and len(ar2) == 1:
        y = ar2[0]
    elif isinstance(ar2, np.ndarray) or isinstance(ar1, block_vector.BlockVector):
        y = ar2
    else:
        raise RuntimeError(
            'ar2 type not recognized. Needs to be np.ndarray or BlockVector'
        )

    if isinstance(x, block_vector.BlockVector) and isinstance(
        y, block_vector.BlockVector
    ):
        assert x.nblocks == y.nblocks, "Number of blocks does not match"
        assert (
            not x.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
        assert (
            not y.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'

        res = block_vector.BlockVector(x.nblocks)
        for i in range(x.nblocks):
            res.set_block(
                i,
                setdiff1d(x.get_block(i), y.get_block(i), assume_unique=assume_unique),
            )
        return res
    elif isinstance(x, block_vector.BlockVector) and isinstance(y, np.ndarray):
        assert (
            not x.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'
        res = block_vector.BlockVector(x.nblocks)
        for i in range(x.nblocks):
            res.set_block(
                i, np.setdiff1d(x.get_block(i), y, assume_unique=assume_unique)
            )
        return res
    elif isinstance(x, np.ndarray) and isinstance(y, block_vector.BlockVector):
        assert (
            not y.has_none
        ), 'Operation not allowed with None blocks. Specify all blocks in BlockVector'

        res = block_vector.BlockVector(y.nblocks)
        for i in range(y.nblocks):
            res.set_block(
                i, np.setdiff1d(x, y.get_block(i), assume_unique=assume_unique)
            )
        return res
    else:
        return np.setdiff1d(x, y, assume_unique=assume_unique)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
This module contains a function to generate a list of the Vars which appear
in expressions in the active tree. Note this is not the same as
component_data_objects(Var) because it does not look for Var objects which are
not used in any expressions and it does not care if the Vars it finds are
actually in the subtree or not.
"""

from pyomo.core import Block
from pyomo.core.expr.visitor import IdentifyVariableVisitor


def get_vars_from_components(
    block,
    ctype,
    include_fixed=True,
    active=None,
    sort=False,
    descend_into=Block,
    descent_order=None,
):
    """Returns a generator of all the Var objects which are used in Constraint
    expressions on the block. By default, this recurses into sub-blocks.

    Args:
        ctype: The type of component from which to get Vars, assumed to have
               an expr attribute.
        include_fixed: Whether or not to include fixed variables
        active: Whether to find Vars that appear in Constraints accessible
                via the active tree
        sort: sort method for iterating through Constraint objects
        descend_into: Ctypes to descend into when finding Constraints
        descent_order: Traversal strategy for finding the objects of type ctype
    """
    visitor = IdentifyVariableVisitor(include_fixed, {})
    seen = set()
    for constraint in block.component_data_objects(
        ctype,
        active=active,
        sort=sort,
        descend_into=descend_into,
        descent_order=descent_order,
    ):
        for var in visitor.walk_expression(constraint.expr):
            if id(var) not in seen:
                seen.add(id(var))
                yield var



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import numpy as np
import abc
from pyomo.contrib.pynumero.interfaces.cyipopt_interface import CyIpoptProblemInterface
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.pynumero.sparse.block_vector import BlockVector
from pyomo.environ import Var, Constraint, value
from pyomo.core.base.var import VarData
from pyomo.common.modeling import unique_component_name

"""
This module is used for interfacing a multi-input / multi-output external
evaluation code with a Pyomo model and then solve the coupled model
with CyIpopt.

To use this interface:
   * inherit from ExternalInputOutputModel and implement the necessary methods
     (This provides methods to set the input values, evaluate the output values,
     and evaluate the jacobian of the outputs with respect to the inputs.)
   * create a PyomoExternalCyIpoptProblem object, giving it your pyomo model, an
     instance of the derived ExternalInputOutputModel, a list of the Pyomo variables
     that map to the inputs of the external model, and a list of the Pyomo variables
     that map to the outputs from the external model.
   * The standard CyIpopt solver interface can be called using the PyomoExternalCyIpoptProblem

See the PyNumero tests for this interface to see an example of use.

Todo:
   * Currently, you cannot "fix" a pyomo variable that corresponds to an input or output
     and you must use a constraint instead (this is because Pyomo removes fixed variables
     before sending them to the solver)
   * Remove the dummy variable and constraint once Pyomo supports non-removal of certain
     variables
"""


class ExternalInputOutputModel(object, metaclass=abc.ABCMeta):
    """
    This is the base class for building external input output models
    for use with Pyomo and CyIpopt
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def set_inputs(self, input_values):
        """
        This method is called by the solver to set the current values
        for the input variables. The derived class must cache these if
        necessary for any subsequent calls to evaluate_outputs or
        evaluate_derivatives.
        """
        pass

    @abc.abstractmethod
    def evaluate_outputs(self):
        """
        Compute the outputs from the model (using the values
        set in input_values) and return as a numpy array
        """
        pass

    @abc.abstractmethod
    def evaluate_derivatives(self):
        """
        Compute the derivatives of the outputs with respect
        to the inputs (using the values set in input_values).
        This should be a dense matrix with the rows in
        the order of the output variables and the cols in
        the order of the input variables.
        """
        pass

    # ToDo: Hessians not yet handled


class PyomoExternalCyIpoptProblem(CyIpoptProblemInterface):
    def __init__(
        self,
        pyomo_model,
        ex_input_output_model,
        inputs,
        outputs,
        outputs_eqn_scaling=None,
        nl_file_options=None,
    ):
        """
        Create an instance of this class to pass as a problem to CyIpopt.

        Parameters
        ----------
        pyomo_model : ConcreteModel
           The ConcreteModel representing the Pyomo part of the problem. This
           model must contain Pyomo variables for the inputs and the outputs.

        ex_input_output_model : ExternalInputOutputModel
           An instance of a derived class (from ExternalInputOutputModel) that provides
           the methods to compute the outputs and the derivatives.

        inputs : list of Pyomo variables (VarData)
           The Pyomo model needs to have variables to represent the inputs to the
           external model. This is the list of those input variables in the order
           that corresponds to the input_values vector provided in the set_inputs call.

        outputs : list of Pyomo variables (VarData)
          The Pyomo model needs to have variables to represent the outputs from the
          external model. This is the list of those output variables in the order
          that corresponds to the numpy array returned from the evaluate_outputs call.

        outputs_eqn_scaling : list or array-like or None
          This sets the value of scaling parameters for the additional
          output equations that are generated. No scaling is done if this
          is set to None.
        """
        self._pyomo_model = pyomo_model
        self._ex_io_model = ex_input_output_model

        # verify that the inputs and outputs were passed correctly
        self._inputs = [v for v in inputs]
        for v in self._inputs:
            if not isinstance(v, VarData):
                raise RuntimeError(
                    'Argument inputs passed to PyomoExternalCyIpoptProblem must be'
                    ' a list of VarData objects. Note: if you have an indexed variable, pass'
                    ' each index as a separate entry in the list (e.g., inputs=[m.x[1], m.x[2]]).'
                )

        self._outputs = [v for v in outputs]
        for v in self._outputs:
            if not isinstance(v, VarData):
                raise RuntimeError(
                    'Argument outputs passed to PyomoExternalCyIpoptProblem must be'
                    ' a list of VarData objects. Note: if you have an indexed variable, pass'
                    ' each index as a separate entry in the list (e.g., inputs=[m.x[1], m.x[2]]).'
                )

        # we need to add a dummy variable and constraint to the pyomo_nlp
        # to make sure it does not remove variables that do not
        # appear in the pyomo part of the model - also ensure unique name in case model
        # is used in more than one instance of this class
        # ToDo: Improve this by convincing Pyomo not to remove the inputs and outputs
        dummy_var_name = unique_component_name(
            self._pyomo_model, '_dummy_variable_CyIpoptPyomoExNLP'
        )
        dummy_var = Var()
        setattr(self._pyomo_model, dummy_var_name, dummy_var)
        dummy_con_name = unique_component_name(
            self._pyomo_model, '_dummy_constraint_CyIpoptPyomoExNLP'
        )
        dummy_con = Constraint(
            expr=getattr(self._pyomo_model, dummy_var_name)
            == sum(v for v in self._inputs) + sum(v for v in self._outputs)
        )
        setattr(self._pyomo_model, dummy_con_name, dummy_con)

        # initialize the dummy var to the right hand side
        dummy_var_value = 0
        for v in self._inputs:
            if v.value is not None:
                dummy_var_value += value(v)
        for v in self._outputs:
            if v.value is not None:
                dummy_var_value += value(v)
        dummy_var.value = dummy_var_value

        # make an nlp interface from the pyomo model
        self._pyomo_nlp = PyomoNLP(self._pyomo_model, nl_file_options)

        # create initial value vectors for primals and duals
        init_primals = self._pyomo_nlp.init_primals()
        init_duals_pyomo = self._pyomo_nlp.init_duals()
        if np.any(np.isnan(init_duals_pyomo)):
            # set initial values to 1 for any entries that we don't get
            # (typically, all are set, or none are set)
            init_duals_pyomo[np.isnan(init_duals_pyomo)] = 1.0
        init_duals_ex = np.ones(len(self._outputs), dtype=np.float64)
        init_duals = BlockVector(2)
        init_duals.set_block(0, init_duals_pyomo)
        init_duals.set_block(1, init_duals_ex)

        # build the map from inputs and outputs to the full x vector
        self._input_columns = self._pyomo_nlp.get_primal_indices(self._inputs)
        # self._input_x_mask = np.zeros(self._pyomo_nlp.n_primals(), dtype=np.float64)
        # self._input_x_mask[self._input_columns] = 1.0
        self._output_columns = self._pyomo_nlp.get_primal_indices(self._outputs)
        # self._output_x_mask = np.zeros(self._pyomo_nlp.n_primals(), dtype=np.float64)
        # self._output_x_mask[self._output_columns] = 1.0

        # create caches for primals and duals
        self._cached_primals = init_primals.copy()
        self._cached_duals = init_duals.clone(copy=True)
        self._cached_obj_factor = 1.0

        # set the initial values for the pyomo primals and duals
        self._pyomo_nlp.set_primals(self._cached_primals)
        self._pyomo_nlp.set_duals(self._cached_duals.get_block(0))
        # set the initial values for the external inputs
        ex_inputs = self._ex_io_inputs_from_full_primals(self._cached_primals)
        self._ex_io_model.set_inputs(ex_inputs)

        # create the lower and upper bounds for the complete problem
        pyomo_nlp_con_lb = self._pyomo_nlp.constraints_lb()
        ex_con_lb = np.zeros(len(self._outputs), dtype=np.float64)
        self._gL = np.concatenate((pyomo_nlp_con_lb, ex_con_lb))
        pyomo_nlp_con_ub = self._pyomo_nlp.constraints_ub()
        ex_con_ub = np.zeros(len(self._outputs), dtype=np.float64)
        self._gU = np.concatenate((pyomo_nlp_con_ub, ex_con_ub))

        # create the scaling parameters if they are provided
        self._obj_scaling = self._pyomo_nlp.get_obj_scaling()
        self._primals_scaling = self._pyomo_nlp.get_primals_scaling()
        pyomo_constraints_scaling = self._pyomo_nlp.get_constraints_scaling()
        self._constraints_scaling = None
        # check if we need constraint scaling, and if so, add in the
        # outputs_eqn_scaling
        if pyomo_constraints_scaling is not None or outputs_eqn_scaling is not None:
            if pyomo_constraints_scaling is None:
                pyomo_constraints_scaling = np.ones(
                    self._pyomo_nlp.n_primals(), dtype=np.float64
                )
            if outputs_eqn_scaling is None:
                outputs_eqn_scaling = np.ones(len(self._outputs), dtype=np.float64)
            if type(outputs_eqn_scaling) is list:
                outputs_eqn_scaling = np.asarray(outputs_eqn_scaling, dtype=np.float64)
            self._constraints_scaling = np.concatenate(
                (pyomo_constraints_scaling, outputs_eqn_scaling)
            )

        ### setup the jacobian structures
        self._jac_pyomo = self._pyomo_nlp.evaluate_jacobian()

        # We will be mapping the dense external jacobian (doutputs/dinputs)
        # to the correct columns from the full x vector
        ex_start_row = self._pyomo_nlp.n_constraints()

        jac_ex = self._ex_io_model.evaluate_derivatives()

        # the jacobian returned from the external model is in the
        # space of the external model only. We need to shift
        # the rows down and shift the columns appropriately
        jac_ex_irows = np.copy(jac_ex.row)
        jac_ex_irows += ex_start_row
        jac_ex_jcols = np.copy(jac_ex.col)
        for z, col in enumerate(jac_ex_jcols):
            jac_ex_jcols[z] = self._input_columns[col]
        jac_ex_data = np.copy(jac_ex.data)

        # CDL: this code was for the dense version of evaluate_derivatives
        # for i in range(len(self._outputs)):
        #     for j in range(len(self._inputs)):
        #         jac_ex_irows.append(ex_start_row + i)
        #         jac_ex_jcols.append(self._input_columns[j])
        #         jac_ex_data.append(jac_ex[i,j])

        jac_ex_output_irows = list()
        jac_ex_output_jcols = list()
        jac_ex_output_data = list()

        # add the jac for output variables from the extra equations
        for i in range(len(self._outputs)):
            jac_ex_output_irows.append(ex_start_row + i)
            jac_ex_output_jcols.append(self._output_columns[i])
            jac_ex_output_data.append(-1.0)

        self._full_jac_irows = np.concatenate(
            (self._jac_pyomo.row, jac_ex_irows, jac_ex_output_irows)
        )
        self._full_jac_jcols = np.concatenate(
            (self._jac_pyomo.col, jac_ex_jcols, jac_ex_output_jcols)
        )
        self._full_jac_data = np.concatenate(
            (self._jac_pyomo.data, jac_ex_data, jac_ex_output_data)
        )

        # currently, this interface does not do anything with Hessians

        # Call CyIpoptProblemInterface.__init__, which calls
        # cyipopt.Problem.__init__
        super(PyomoExternalCyIpoptProblem, self).__init__()

    def load_x_into_pyomo(self, primals):
        """
        Use this method to load a numpy array of values into the corresponding
        Pyomo variables (e.g., the solution from CyIpopt)

        Parameters
        ----------
        primals : numpy array
           The array of values that will be given to the Pyomo variables. The
           order of this array is the same as the order in the PyomoNLP created
           internally.
        """
        pyomo_variables = self._pyomo_nlp.get_pyomo_variables()
        for i, v in enumerate(primals):
            pyomo_variables[i].set_value(v)

    def _set_primals_if_necessary(self, primals):
        if not np.array_equal(primals, self._cached_primals):
            self._pyomo_nlp.set_primals(primals)
            ex_inputs = self._ex_io_inputs_from_full_primals(primals)
            self._ex_io_model.set_inputs(ex_inputs)
            self._cached_primals = primals.copy()

    def _set_duals_if_necessary(self, duals):
        if not np.array_equal(duals, self._cached_duals):
            self._cached_duals.copy_from(duals)
            self._pyomo_nlp.set_duals(self._cached_duals.get_block(0))

    def _set_obj_factor_if_necessary(self, obj_factor):
        if obj_factor != self._cached_obj_factor:
            self._pyomo_nlp.set_obj_factor(obj_factor)
            self._cached_obj_factor = obj_factor

    def x_init(self):
        return self._pyomo_nlp.init_primals()

    def x_lb(self):
        return self._pyomo_nlp.primals_lb()

    def x_ub(self):
        return self._pyomo_nlp.primals_ub()

    def g_lb(self):
        return self._gL.copy()

    def g_ub(self):
        return self._gU.copy()

    def scaling_factors(self):
        return self._obj_scaling, self._primals_scaling, self._constraints_scaling

    def objective(self, primals):
        self._set_primals_if_necessary(primals)
        return self._pyomo_nlp.evaluate_objective()

    def gradient(self, primals):
        self._set_primals_if_necessary(primals)
        return self._pyomo_nlp.evaluate_grad_objective()

    def constraints(self, primals):
        self._set_primals_if_necessary(primals)
        pyomo_constraints = self._pyomo_nlp.evaluate_constraints()
        ex_io_outputs = self._ex_io_model.evaluate_outputs()
        ex_io_constraints = ex_io_outputs - self._ex_io_outputs_from_full_primals(
            primals
        )
        constraints = BlockVector(2)
        constraints.set_block(0, pyomo_constraints)
        constraints.set_block(1, ex_io_constraints)
        return constraints.flatten()

    def jacobianstructure(self):
        return self._full_jac_irows, self._full_jac_jcols

    def jacobian(self, primals):
        self._set_primals_if_necessary(primals)
        self._pyomo_nlp.evaluate_jacobian(out=self._jac_pyomo)
        pyomo_data = self._jac_pyomo.data
        ex_io_deriv = self._ex_io_model.evaluate_derivatives()
        # CDL: dense version: ex_io_deriv = self._ex_io_model.evaluate_derivatives().flatten('C')
        self._full_jac_data[0 : len(pyomo_data)] = pyomo_data
        self._full_jac_data[
            len(pyomo_data) : len(pyomo_data) + len(ex_io_deriv.data)
        ] = ex_io_deriv.data
        # CDL: dense version: self._full_jac_data[len(pyomo_data):len(pyomo_data)+len(ex_io_deriv)] = ex_io_deriv

        # the -1s for the output variables should still be  here
        return self._full_jac_data

    def hessianstructure(self):
        return np.zeros(0), np.zeros(0)
        # raise NotImplementedError('No Hessians for now')

    def hessian(self, x, y, obj_factor):
        raise NotImplementedError('No Hessians for now')

    def _ex_io_inputs_from_full_primals(self, primals):
        return primals[self._input_columns]
        # return np.compress(self._input_x_mask, primals)

    def _ex_io_outputs_from_full_primals(self, primals):
        return primals[self._output_columns]
        # return  np.compress(self._output_x_mask, primals)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
PLY: Python Lex-Yacc
====================

This module is derived from the PLY package (version 3.11) by David Beazley.

It has been modified to disallow loading parse tables from arbitrary
files or pickles.  This simplifies management of the parse table module
and avoids potential security concerns.

The original PLY project (https://github.com/dabeaz/ply) was archived on
December 21, 2025 and is no longer developed or maintained.

PLY (3.11) documentation is available here:  https://www.dabeaz.com/ply/ply.html

"""

__version__ = '3.11.post1'
__all__ = ['lex', 'yacc']



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import itertools
import re

from difflib import SequenceMatcher, unified_diff

from pyomo.repn.tests.diffutils import compare_floats, load_baseline
from pyomo.repn.ampl import TextNLDebugTemplate as template

_norm_whitespace = re.compile(r'[^\S\n]+')
_norm_integers = re.compile(r'(?m)\.0+$')
_norm_comment = re.compile(r'\s*#\s*')
_strip_comment = re.compile(r'\s*#.*')
_norm_negation = re.compile(r'(?m)^o2(\s*#\s*\*)?\nn-1(.0)?\s*\n')
_norm_timesone = re.compile(r'(?m)^o2(\s*#\s*\*)?\nn1(.0)?\s*\n')
_norm_double_negation = re.compile(r'(?m)^o16(\s*#\s*-)?\no16(\s*#\s*-)?\n')


def _update_subsets(subset, base, test):
    for i, j in zip(*subset):
        # Try checking for numbers
        if base[i][0] == 'n' and test[j][0] == 'n':
            if compare_floats(base[i][1:], test[j][1:]):
                test[j] = base[i]
        elif compare_floats(base[i], test[j]):
            test[j] = base[i]
        else:
            # try stripping comments, but only if it results in a match
            base_nc = _strip_comment.sub('', base[i])
            test_nc = _strip_comment.sub('', test[j])
            if compare_floats(base_nc, test_nc):
                if len(base_nc) > len(test_nc):
                    test[j] = base[i]
                else:
                    base[i] = test[j]


def _preprocess_data(data):
    # Normalize negation (convert " * -1" to the negation operator)
    data = _norm_negation.sub(template.negation, data)
    # Normalize double negation (convert "-(-x)" to x)
    data = _norm_double_negation.sub('', data)
    # Remove multiplication by 1
    data = _norm_timesone.sub('', data)
    # Normalize consecutive whitespace to a single space
    data = _norm_whitespace.sub(' ', data)
    # preface all comments with a single tab character
    data = _norm_comment.sub('\t#', data)
    # Normalize floating point integers to integers
    data = _norm_integers.sub('', data)
    # return the sequence of lines
    return data.splitlines()


def nl_diff(base, test, baseline='baseline', testfile='testfile'):
    if test == base:
        return [], []

    test = _preprocess_data(test)
    base = _preprocess_data(base)
    if test == base:
        return [], []

    # First do a quick pass to check / standardize embedded numbers.
    # This is a little fragile (it requires that the embedded constants
    # appear in the same order in the base and test files), but we see
    # cases where differences within numerical tolerances lead to huge
    # add / delete chunks (instead of small replace chunks) from the
    # SequenceMatcher (because it is not as fast / aggressive as Unix
    # diff).  Those add/remove chunks are ignored by the _update_subsets
    # code below, leading to unnecessary test failures.
    test_nlines = list(x for x in enumerate(test) if x[1] and x[1][0] == 'n')
    base_nlines = list(x for x in enumerate(base) if x[1] and x[1][0] == 'n')
    if len(test_nlines) == len(base_nlines):
        for t_line, b_line in zip(test_nlines, base_nlines):
            if compare_floats(t_line[1][1:], b_line[1][1:]):
                test[t_line[0]] = base[b_line[0]]

    for group in SequenceMatcher(None, base, test).get_grouped_opcodes(3):
        for tag, i1, i2, j1, j2 in group:
            if tag != 'replace':
                continue
            _update_subsets((range(i1, i2), range(j1, j2)), base, test)

    if test == base:
        return [], []

    print(
        ''.join(
            unified_diff(
                [_ + "\n" for _ in base],
                [_ + "\n" for _ in test],
                fromfile=baseline,
                tofile=testfile,
            )
        )
    )
    return base, test


def load_nl_baseline(baseline, testfile, version='nl'):
    return load_baseline(baseline, testfile, 'nl', version)


def load_and_compare_nl_baseline(baseline, testfile, version='nl'):
    return nl_diff(*load_nl_baseline(baseline, testfile, version))


if __name__ == '__main__':
    import sys

    base, test = load_and_compare_nl_baseline(sys.argv[1], sys.argv[2])
    sys.exit(1 if base or test else 0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.opt import check_optimal_termination
from pyomo.common.dependencies import attempt_import
from .interface import InteriorPointInterface
from .linalg.scipy_interface import ScipyInterface

np, numpy_available = attempt_import(
    'numpy', 'Interior point requires numpy', minimum_version='1.13.0'
)


# Todo: This function currently used IPOPT for the initial solve - should accept solver
def inv_reduced_hessian_barrier(
    model, independent_variables, bound_tolerance=1e-6, solver_options=None, tee=False
):
    """
    This function computes the inverse of the reduced Hessian of a problem at the
    solution. This function first solves the problem with Ipopt and then generates
    the KKT system for the barrier subproblem to compute the inverse reduced hessian.

    For more information on the reduced Hessian, see "Numerical Optimization", 2nd Edition
    Nocedal and Wright, 2006.

    The approach used in this method can be found in, "Computational Strategies for
    the Optimal Operation of Large-Scale Chemical Processes", Dissertation, V. Zavala
    2008. See section 3.2.1.

    Parameters
    ----------
    model : Pyomo model
        The Pyomo model that we want to solve and analyze
    independent_variables : list of Pyomo variables
        This is the list of independent variables for computing the reduced hessian.
        These variables must not be at their bounds at the solution of the
        optimization problem.
    bound_tolerance : float
       The tolerance to use when checking if the variables are too close to their bound.
       If they are too close, then the routine will exit without a reduced hessian.
    solver_options: dictionary
        Additional solver options to consider.
    tee : bool
       This flag is sent to the tee option of the solver. If true, then the solver
       log is output to the console.
    """
    m = model

    # make sure the necessary suffixes are added
    # so the reduced hessian kkt system is setup correctly from
    # the ipopt solution
    if not hasattr(m, 'ipopt_zL_out'):
        m.ipopt_zL_out = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    if not hasattr(m, 'ipopt_zU_out'):
        m.ipopt_zU_out = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    if not hasattr(m, 'ipopt_zL_in'):
        m.ipopt_zL_in = pyo.Suffix(direction=pyo.Suffix.EXPORT)
    if not hasattr(m, 'ipopt_zU_in'):
        m.ipopt_zU_in = pyo.Suffix(direction=pyo.Suffix.EXPORT)
    if not hasattr(m, 'dual'):
        m.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT_EXPORT)

    # create the ipopt solver
    solver = pyo.SolverFactory('ipopt')

    # copy additional solver options
    if solver_options is not None:
        for key in solver_options:
            solver.options[key] = solver_options[key]

    # set options to prevent bounds relaxation (and 0 slacks)
    solver.options['bound_relax_factor'] = 0
    solver.options['honor_original_bounds'] = 'no'

    # solve the problem
    status = solver.solve(m, tee=tee)
    if not check_optimal_termination(status):
        return status, None

    # compute the barrier parameter
    # ToDo: this needs to eventually come from the solver itself
    estimated_mu = list()
    for v in m.ipopt_zL_out:
        if v.has_lb():
            estimated_mu.append((pyo.value(v) - v.lb) * m.ipopt_zL_out[v])
    for v in m.ipopt_zU_out:
        if v.has_ub():
            estimated_mu.append((v.ub - pyo.value(v)) * m.ipopt_zU_out[v])
    if len(estimated_mu) == 0:
        mu = 10**-8.6
    else:
        mu = sum(estimated_mu) / len(estimated_mu)
        # check to make sure these estimates were all reasonable
        if any([abs(mu - estmu) > 1e-7 for estmu in estimated_mu]):
            print(
                'Warning: estimated values of mu do not seem consistent - using mu=10^(-8.6)'
            )
            mu = 10**-8.6

    # collect the list of var data objects for the independent variables
    ind_vardatas = list()
    for v in independent_variables:
        if v.is_indexed():
            for k in v:
                ind_vardatas.append(v[k])
        else:
            ind_vardatas.append(v)

    # check that none of the independent variables are at their bounds
    for v in ind_vardatas:
        if (v.has_lb() and pyo.value(v) - v.lb <= bound_tolerance) or (
            v.has_ub() and v.ub - pyo.value(v) <= bound_tolerance
        ):
            raise ValueError(
                "Independent variable: {} has a solution value that is near"
                " its bound (according to tolerance). The reduced hessian"
                " computation does not support this at this time. All"
                " independent variables should be in their interior.".format(v)
            )

    # find the list of indices that we need to make up the reduced hessian
    kkt_builder = InteriorPointInterface(m)
    pyomo_nlp = kkt_builder.pyomo_nlp()
    ind_var_indices = pyomo_nlp.get_primal_indices(ind_vardatas)

    # setup the computation of the reduced hessian
    kkt_builder.set_barrier_parameter(mu)
    kkt = kkt_builder.evaluate_primal_dual_kkt_matrix()
    linear_solver = ScipyInterface(compute_inertia=False)
    linear_solver.do_symbolic_factorization(kkt)
    linear_solver.do_numeric_factorization(kkt)

    n_rh = len(ind_var_indices)
    rhs = np.zeros(kkt.shape[0])
    inv_red_hess = np.zeros((n_rh, n_rh))

    for rhi, vari in enumerate(ind_var_indices):
        rhs[vari] = 1
        v, res = linear_solver.do_back_solve(rhs)
        rhs[vari] = 0
        for rhj, varj in enumerate(ind_var_indices):
            inv_red_hess[rhi, rhj] = v[varj]

    return status, inv_red_hess



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
import logging
import math
import sys
from typing import Optional, List, Dict

from pyomo.contrib.appsi.base import (
    PersistentSolver,
    Results,
    TerminationCondition,
    MIPSolverConfig,
    PersistentBase,
    PersistentSolutionLoader,
)
from pyomo.contrib.appsi.cmodel import cmodel, cmodel_available
from pyomo.common.collections import ComponentMap
from pyomo.common.config import (
    ConfigValue,
    ConfigDict,
    NonNegativeInt,
    NonNegativeFloat,
)
from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import PyomoException
from pyomo.common.log import LogStream
from pyomo.common.tee import capture_output, TeeStream
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.base import SymbolMap, NumericLabeler, TextLabeler
from pyomo.core.base.constraint import _GeneralConstraintData
from pyomo.core.base.expression import ScalarExpression
from pyomo.core.base.param import _ParamData
from pyomo.core.base.sos import _SOSConstraintData
from pyomo.core.base.var import Var, ScalarVar, _GeneralVarData
import pyomo.core.expr.expr_common as common
import pyomo.core.expr as EXPR
from pyomo.core.expr.numvalue import (
    value,
    is_constant,
    is_fixed,
    native_numeric_types,
    native_types,
    nonpyomo_leaf_types,
)
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn.util import valid_expr_ctypes_minlp

logger = logging.getLogger(__name__)
MaingoVar = namedtuple("MaingoVar", "type name lb ub init")
maingopy, maingopy_available = attempt_import("maingopy")
# Note that importing maingo_solvermodel will trigger the import of
# maingopy, so we defer that import using attempt_import (which will
# always succeed, even if maingopy is not available)
maingo_solvermodel = attempt_import("pyomo.contrib.appsi.solvers.maingo_solvermodel")[0]


class MAiNGOConfig(MIPSolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super(MAiNGOConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )
        self.tolerances: ConfigDict = self.declare(
            'tolerances', ConfigDict(implicit=True)
        )

        self.tolerances.epsilonA: Optional[float] = self.tolerances.declare(
            'epsilonA',
            ConfigValue(
                domain=NonNegativeFloat,
                default=1e-5,
                description="Absolute optimality tolerance",
            ),
        )
        self.tolerances.epsilonR: Optional[float] = self.tolerances.declare(
            'epsilonR',
            ConfigValue(
                domain=NonNegativeFloat,
                default=1e-5,
                description="Relative optimality tolerance",
            ),
        )
        self.tolerances.deltaEq: Optional[float] = self.tolerances.declare(
            'deltaEq',
            ConfigValue(
                domain=NonNegativeFloat, default=1e-6, description="Equality tolerance"
            ),
        )

        self.tolerances.deltaIneq: Optional[float] = self.tolerances.declare(
            'deltaIneq',
            ConfigValue(
                domain=NonNegativeFloat,
                default=1e-6,
                description="Inequality tolerance",
            ),
        )
        self.declare("logfile", ConfigValue(domain=str, default=""))
        self.declare("solver_output_logger", ConfigValue(default=logger))
        self.declare(
            "log_level", ConfigValue(domain=NonNegativeInt, default=logging.INFO)
        )


class MAiNGOSolutionLoader(PersistentSolutionLoader):
    def load_vars(self, vars_to_load=None):
        self._assert_solution_still_valid()
        self._solver.load_vars(vars_to_load=vars_to_load)

    def get_primals(self, vars_to_load=None):
        self._assert_solution_still_valid()
        return self._solver.get_primals(vars_to_load=vars_to_load)


class MAiNGOResults(Results):
    def __init__(self, solver):
        super(MAiNGOResults, self).__init__()
        self.wallclock_time = None
        self.cpu_time = None
        self.globally_optimal = None
        self.solution_loader = MAiNGOSolutionLoader(solver=solver)


class MAiNGO(PersistentBase, PersistentSolver):
    """
    Interface to MAiNGO
    """

    _available = None

    def __init__(self, only_child_vars=False):
        super(MAiNGO, self).__init__(only_child_vars=only_child_vars)
        self._config = MAiNGOConfig()
        self._solver_options = dict()
        self._solver_model = None
        self._mymaingo = None
        self._symbol_map = SymbolMap()
        self._labeler = None
        self._maingo_vars = []
        self._objective = None
        self._cons = []
        self._pyomo_var_to_solver_var_id_map = dict()
        self._last_results_object: Optional[MAiNGOResults] = None

    def available(self):
        if self._available is None:
            if maingopy_available:
                MAiNGO._available = True
            else:
                MAiNGO._available = MAiNGO.Availability.NotFound
        return self._available

    def version(self):
        import importlib.metadata

        try:
            version = importlib.metadata.version('maingopy').split('.')
        except ImportError:
            return None
        for i, n in enumerate(version):
            try:
                version[i] = int(version[i])
            except:
                pass
        return tuple(version)

    @property
    def config(self) -> MAiNGOConfig:
        return self._config

    @config.setter
    def config(self, val: MAiNGOConfig):
        self._config = val

    @property
    def maingo_options(self):
        """
        A dictionary mapping solver options to values for those options. These
        are solver specific.

        Returns
        -------
        dict
            A dictionary mapping solver options to values for those options
        """
        return self._solver_options

    @maingo_options.setter
    def maingo_options(self, val: Dict):
        self._solver_options = val

    @property
    def symbol_map(self):
        return self._symbol_map

    def _solve(self, timer: HierarchicalTimer):
        ostreams = [
            LogStream(
                level=self.config.log_level, logger=self.config.solver_output_logger
            )
        ]
        if self.config.stream_solver:
            ostreams.append(sys.stdout)

        with capture_output(output=TeeStream(*ostreams), capture_fd=False):
            config = self.config
            options = self.maingo_options

            self._mymaingo = maingopy.MAiNGO(self._solver_model)

            self._mymaingo.set_option("loggingDestination", 2)
            self._mymaingo.set_log_file_name(config.logfile)
            self._mymaingo.set_option("epsilonA", config.tolerances.epsilonA)
            self._mymaingo.set_option("epsilonR", config.tolerances.epsilonR)
            self._mymaingo.set_option("deltaEq", config.tolerances.deltaEq)
            self._mymaingo.set_option("deltaIneq", config.tolerances.deltaIneq)

            if config.time_limit is not None:
                self._mymaingo.set_option("maxTime", config.time_limit)
            if config.mip_gap is not None:
                self._mymaingo.set_option("epsilonR", config.mip_gap)
            for key, option in options.items():
                self._mymaingo.set_option(key, option)

            timer.start("MAiNGO solve")
            self._mymaingo.solve()
            timer.stop("MAiNGO solve")

        return self._postsolve(timer)

    def solve(self, model, timer: HierarchicalTimer = None):
        StaleFlagManager.mark_all_as_stale()

        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if timer is None:
            timer = HierarchicalTimer()
        if model is not self._model:
            timer.start("set_instance")
            self.set_instance(model)
            timer.stop("set_instance")
        else:
            timer.start("Update")
            self.update(timer=timer)
            timer.stop("Update")
        res = self._solve(timer)
        self._last_results_object = res
        if self.config.report_timing:
            logger.info("\n" + str(timer))
        return res

    def _process_domain_and_bounds(self, var):
        _v, _lb, _ub, _fixed, _domain_interval, _value = self._vars[id(var)]
        lb, ub, step = _domain_interval

        if _fixed:
            lb = _value
            ub = _value
        else:
            if lb is None and _lb is None:
                logger.warning(
                    "No lower bound for variable "
                    + var.getname()
                    + " set. Using -1e10 instead. Please consider setting a valid lower bound."
                )
            if ub is None and _ub is None:
                logger.warning(
                    "No upper bound for variable "
                    + var.getname()
                    + " set. Using +1e10 instead. Please consider setting a valid upper bound."
                )

            if _lb is None:
                _lb = -1e10
            if _ub is None:
                _ub = 1e10
            if lb is None:
                lb = -1e10
            if ub is None:
                ub = 1e10

            lb = max(value(_lb), lb)
            ub = min(value(_ub), ub)

        if step == 0:
            vtype = maingopy.VT_CONTINUOUS
        elif step == 1:
            if lb == 0 and ub == 1:
                vtype = maingopy.VT_BINARY
            else:
                vtype = maingopy.VT_INTEGER
        else:
            raise ValueError(
                f"Unrecognized domain step: {step} (should be either 0 or 1)"
            )

        return lb, ub, vtype

    def _add_variables(self, variables: List[_GeneralVarData]):
        for var in variables:
            varname = self._symbol_map.getSymbol(var, self._labeler)
            lb, ub, vtype = self._process_domain_and_bounds(var)
            self._maingo_vars.append(
                MaingoVar(name=varname, type=vtype, lb=lb, ub=ub, init=var.value)
            )
            self._pyomo_var_to_solver_var_id_map[id(var)] = len(self._maingo_vars) - 1

    def _add_params(self, params: List[_ParamData]):
        pass

    def _reinit(self):
        saved_config = self.config
        saved_options = self.maingo_options
        saved_update_config = self.update_config
        self.__init__(only_child_vars=self._only_child_vars)
        self.config = saved_config
        self.maingo_options = saved_options
        self.update_config = saved_update_config

    def set_instance(self, model):
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if not self.available():
            c = self.__class__
            raise PyomoException(
                f"Solver {c.__module__}.{c.__qualname__} is not available "
                f"({self.available()})."
            )
        self._reinit()
        self._model = model
        if self.use_extensions and cmodel_available:
            self._expr_types = cmodel.PyomoExprTypes()

        if self.config.symbolic_solver_labels:
            self._labeler = TextLabeler()
        else:
            self._labeler = NumericLabeler("x")

        self.add_block(model)

        self._solver_model = maingo_solvermodel.SolverModel(
            var_list=self._maingo_vars,
            con_list=self._cons,
            objective=self._objective,
            idmap=self._pyomo_var_to_solver_var_id_map,
            logger=logger,
        )

    def _add_constraints(self, cons: List[_GeneralConstraintData]):
        self._cons += cons

    def _add_sos_constraints(self, cons: List[_SOSConstraintData]):
        if len(cons) >= 1:
            raise NotImplementedError(
                "MAiNGO does not currently support SOS constraints."
            )
        pass

    def _remove_constraints(self, cons: List[_GeneralConstraintData]):
        for con in cons:
            self._cons.remove(con)

    def _remove_sos_constraints(self, cons: List[_SOSConstraintData]):
        if len(cons) >= 1:
            raise NotImplementedError(
                "MAiNGO does not currently support SOS constraints."
            )
        pass

    def _remove_variables(self, variables: List[_GeneralVarData]):
        removed_maingo_vars = []
        for var in variables:
            varname = self._symbol_map.getSymbol(var, self._labeler)
            del self._maingo_vars[self._pyomo_var_to_solver_var_id_map[id(var)]]
            removed_maingo_vars += [self._pyomo_var_to_solver_var_id_map[id(var)]]
            del self._pyomo_var_to_solver_var_id_map[id(var)]

        # Update _pyomo_var_to_solver_var_id_map to account for removed variables
        for pyomo_var, maingo_var_id in self._pyomo_var_to_solver_var_id_map.items():
            num_removed = 0
            for removed_var in removed_maingo_vars:
                if removed_var <= maingo_var_id:
                    num_removed += 1
            self._pyomo_var_to_solver_var_id_map[pyomo_var] = (
                maingo_var_id - num_removed
            )

    def _remove_params(self, params: List[_ParamData]):
        pass

    def _update_variables(self, variables: List[_GeneralVarData]):
        for var in variables:
            if id(var) not in self._pyomo_var_to_solver_var_id_map:
                raise ValueError(
                    'The Var provided to update_var needs to be added first: {0}'.format(
                        var
                    )
                )
            lb, ub, vtype = self._process_domain_and_bounds(var)
            self._maingo_vars[self._pyomo_var_to_solver_var_id_map[id(var)]] = (
                MaingoVar(name=var.name, type=vtype, lb=lb, ub=ub, init=var.value)
            )

    def update_params(self):
        vars = [var[0] for var in self._vars.values()]
        self._update_variables(vars)

    def _set_objective(self, obj):

        if not obj.sense in {minimize, maximize}:
            raise ValueError("Objective sense is not recognized: {0}".format(obj.sense))
        self._objective = obj

    def _postsolve(self, timer: HierarchicalTimer):
        config = self.config

        mprob = self._mymaingo
        status = mprob.get_status()
        results = MAiNGOResults(solver=self)
        results.wallclock_time = mprob.get_wallclock_solution_time()
        results.cpu_time = mprob.get_cpu_solution_time()

        if status in {maingopy.GLOBALLY_OPTIMAL, maingopy.FEASIBLE_POINT}:
            results.termination_condition = TerminationCondition.optimal
            results.globally_optimal = True
            if status == maingopy.FEASIBLE_POINT:
                results.globally_optimal = False
                logger.warning(
                    "MAiNGO found a feasible solution but did not prove its global optimality."
                )
        elif status == maingopy.INFEASIBLE:
            results.termination_condition = TerminationCondition.infeasible
        else:
            results.termination_condition = TerminationCondition.unknown

        results.best_feasible_objective = None
        results.best_objective_bound = None
        if self._objective is not None:
            try:
                if self._objective.sense == maximize:
                    results.best_feasible_objective = -mprob.get_objective_value()
                else:
                    results.best_feasible_objective = mprob.get_objective_value()
            except:
                results.best_feasible_objective = None
            try:
                if self._objective.sense == maximize:
                    results.best_objective_bound = -mprob.get_final_LBD()
                else:
                    results.best_objective_bound = mprob.get_final_LBD()
            except:
                if self._objective.sense == maximize:
                    results.best_objective_bound = math.inf
                else:
                    results.best_objective_bound = -math.inf

            if results.best_feasible_objective is not None and not math.isfinite(
                results.best_feasible_objective
            ):
                results.best_feasible_objective = None

        timer.start("load solution")
        if config.load_solution:
            if results.termination_condition is TerminationCondition.optimal:
                if not results.globally_optimal:
                    logger.warning(
                        "Loading a feasible but suboptimal solution. "
                        "Please set load_solution=False and check "
                        "results.termination_condition and "
                        "results.found_feasible_solution() before loading a solution."
                    )
                self.load_vars()
            else:
                raise RuntimeError(
                    "A feasible solution was not found, so no solution can be loaded."
                    "Please set opt.config.load_solution=False and check "
                    "results.termination_condition and "
                    "results.best_feasible_objective before loading a solution."
                )
        timer.stop("load solution")

        return results

    def load_vars(self, vars_to_load=None):
        for v, val in self.get_primals(vars_to_load=vars_to_load).items():
            v.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    def get_primals(self, vars_to_load=None):
        if not self._mymaingo.get_status() in {
            maingopy.GLOBALLY_OPTIMAL,
            maingopy.FEASIBLE_POINT,
        }:
            raise RuntimeError(
                "Solver does not currently have a valid solution."
                "Please check the termination condition."
            )

        var_id_map = self._pyomo_var_to_solver_var_id_map
        ref_vars = self._referenced_variables
        if vars_to_load is None:
            vars_to_load = var_id_map.keys()
        else:
            vars_to_load = [id(v) for v in vars_to_load]

        maingo_var_ids_to_load = [
            var_id_map[pyomo_var_id] for pyomo_var_id in vars_to_load
        ]

        solution_point = self._mymaingo.get_solution_point()
        vals = [solution_point[var_id] for var_id in maingo_var_ids_to_load]

        res = ComponentMap()
        for var_id, val in zip(vars_to_load, vals):
            using_cons, using_sos, using_obj = ref_vars[var_id]
            if using_cons or using_sos or (using_obj is not None):
                res[self._vars[var_id][0]] = val
        return res

    def get_reduced_costs(self, vars_to_load=None):
        raise ValueError("MAiNGO does not support returning Reduced Costs")

    def get_duals(self, cons_to_load=None):
        raise ValueError("MAiNGO does not support returning Duals")

    def update(self, timer: HierarchicalTimer = None):
        super(MAiNGO, self).update(timer=timer)
        self._solver_model = maingo_solvermodel.SolverModel(
            var_list=self._maingo_vars,
            con_list=self._cons,
            objective=self._objective,
            idmap=self._pyomo_var_to_solver_var_id_map,
            logger=logger,
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from itertools import product

from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept
from pyomo.contrib.multistart.high_conf_stop import should_stop
from pyomo.contrib.multistart.reinit import strategies
from pyomo.environ import (
    ConcreteModel,
    Constraint,
    NonNegativeReals,
    Objective,
    SolverFactory,
    Var,
    maximize,
    sin,
    value,
)


@unittest.skipIf(not SolverFactory('ipopt').available(), "IPOPT not available")
class MultistartTests(unittest.TestCase):
    """
    Due to stochastic nature of the random restarts, these tests just
    demonstrate, that for a small sample, the test will not do worse than the
    standard solver. this is non-exhaustive due to the randomness. Hence all
    asserts are inequalities.
    """

    def test_as_good_as_standard(self):
        standard_model = build_model()
        SolverFactory('ipopt').solve(standard_model)
        standard_objective_value = value(
            next(standard_model.component_data_objects(Objective, active=True))
        )

        fresh_model = build_model()
        multistart_iterations = 10
        test_trials = 10
        for strategy, _ in product(strategies.keys(), range(test_trials)):
            m2 = fresh_model.clone()
            SolverFactory('multistart').solve(
                m2, iterations=multistart_iterations, strategy=strategy
            )
            clone_objective_value = value(
                next(m2.component_data_objects(Objective, active=True))
            )
            self.assertGreaterEqual(
                clone_objective_value, standard_objective_value
            )  # assumes maximization

    def test_as_good_with_HCS_rule(self):
        """test that the high confidence stopping rule with very lenient
        parameters does no worse.
        """
        # initialize model with data
        m = build_model()

        # create ipopt solver
        SolverFactory('ipopt').solve(m)
        for i in range(5):
            m2 = build_model()
            SolverFactory('multistart').solve(
                m2, iterations=-1, stopping_mass=0.99, stopping_delta=0.99
            )
            m_objectives = m.component_data_objects(Objective, active=True)
            m_obj = next(m_objectives, None)
            m2_objectives = m2.component_data_objects(Objective, active=True)
            m2_obj = next(m2_objectives, None)
            # Assert that multistart solver does no worse than standard solver
            self.assertTrue((value(m2_obj.expr)) >= (value(m_obj.expr) - 0.001))
            del m2

    def test_missing_bounds(self):
        m = ConcreteModel()
        m.x = Var(domain=NonNegativeReals)
        m.obj = Objective(expr=m.x)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.multistart', logging.WARNING):
            SolverFactory('multistart').solve(m)
            self.assertIn(
                "Skipping reinitialization of unbounded "
                "variable x with bounds (0, None).",
                output.getvalue().strip(),
            )

    def test_var_value_None(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.obj = Objective(expr=m.x)
        SolverFactory('multistart').solve(m)

    def test_model_infeasible(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.c = Constraint(expr=m.x >= 2)
        m.o = Objective(expr=m.x)
        SolverFactory('multistart').solve(m, iterations=2)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.multistart', logging.WARNING):
            SolverFactory('multistart').solve(m, iterations=-1, HCS_max_iterations=3)
            self.assertIn(
                "High confidence stopping rule was unable to "
                "complete after 3 iterations.",
                output.getvalue().strip(),
            )

    def test_should_stop(self):
        soln = [0] * 149
        self.assertFalse(should_stop(soln, 0.5, 0.5, 0.001))
        soln += [0.001]
        self.assertTrue(should_stop(soln, 0.5, 0.5, 0.001))
        soln = [0] * 149 + [0.01]
        self.assertFalse(should_stop(soln, 0.5, 0.5, 0.001))
        soln = [0] * 149 + [-0.001]
        self.assertTrue(should_stop(soln, 0.5, 0.5, 0.001))

    def test_multiple_obj(self):
        m = ConcreteModel()
        m.x = Var()
        m.o = Objective(expr=m.x)
        m.o2 = Objective(expr=m.x)
        with self.assertRaisesRegex(RuntimeError, "multiple active objectives"):
            SolverFactory('multistart').solve(m)

    def test_no_obj(self):
        m = ConcreteModel()
        m.x = Var()
        with self.assertRaisesRegex(RuntimeError, "no active objective"):
            SolverFactory('multistart').solve(m)

    def test_const_obj(self):
        m = ConcreteModel()
        m.x = Var()
        m.o = Objective(expr=5)
        with self.assertRaisesRegex(RuntimeError, "constant objective"):
            SolverFactory('multistart').solve(m)


def build_model():
    """Simple non-convex model with many local minima"""
    model = ConcreteModel()
    model.x1 = Var(initialize=1, bounds=(0, 100))
    model.x2 = Var(initialize=5, bounds=(5, 6))
    model.x2.fix(5)
    model.objtv = Objective(expr=model.x1 * sin(model.x1), sense=maximize)
    return model


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Problem Writer for GAMS Format Files
#

from io import StringIO

from pyomo.common.gc_manager import PauseGC
import pyomo.core.expr as EXPR
from pyomo.core.expr.numvalue import (
    value,
    as_numeric,
    native_types,
    native_numeric_types,
    nonpyomo_leaf_types,
)
from pyomo.core.expr.visitor import _ToStringVisitor
from pyomo.core.base import (
    SymbolMap,
    ShortNameLabeler,
    NumericLabeler,
    Constraint,
    Objective,
    Var,
    minimize,
    SortComponents,
)
from pyomo.core.base.component import ActiveComponent
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.opt import ProblemFormat
from pyomo.opt.base import AbstractProblemWriter, WriterFactory
from pyomo.repn.util import valid_expr_ctypes_minlp, valid_active_ctypes_minlp, ftoa

import logging

logger = logging.getLogger('pyomo.core')

_legal_unary_functions = {
    'ceil',
    'floor',
    'exp',
    'log',
    'log10',
    'sqrt',
    'sin',
    'cos',
    'tan',
    'asin',
    'acos',
    'atan',
    'sinh',
    'cosh',
    'tanh',
}
_arc_functions = {'acos', 'asin', 'atan'}
_dnlp_functions = {'ceil', 'floor', 'abs'}
_zero_one = {0, 1}
_plusMinusOne = {-1, 1}


def _handle_PowExpression(visitor, node, values):
    # If the exponent is a positive integer, use the power() function.
    # Otherwise, use the ** operator.
    exponent = node.arg(1)
    if exponent.__class__ in native_numeric_types and exponent == int(exponent):
        return f"power({values[0]}, {values[1]})"
    else:
        return f"{values[0]} ** {values[1]}"


def _handle_UnaryFunctionExpression(visitor, node, values):
    if node.name not in _legal_unary_functions:
        raise RuntimeError(
            "GAMS files cannot represent the unary function %s" % (node.name,)
        )
    if node.name in _dnlp_functions:
        visitor.is_discontinuous = True
    if node.name in _arc_functions:
        return f"arc{node.name[1:]}({values[0]})"
    else:
        return node._to_string(values, False, visitor.smap)


def _handle_AbsExpression(visitor, node, values):
    visitor.is_discontinuous = True
    return node._to_string(values, False, visitor.smap)


#
# A visitor pattern that creates a string for an expression
# that is compatible with the GAMS syntax.
#
class ToGamsVisitor(_ToStringVisitor):
    _expression_handlers = {
        EXPR.PowExpression: _handle_PowExpression,
        EXPR.UnaryFunctionExpression: _handle_UnaryFunctionExpression,
        EXPR.AbsExpression: _handle_AbsExpression,
    }

    def __init__(self, smap, treechecker, output_fixed_variables=False):
        super(ToGamsVisitor, self).__init__(False, smap)
        self.treechecker = treechecker
        self.is_discontinuous = False
        self.output_fixed_variables = output_fixed_variables

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in native_types:
            try:
                return True, ftoa(node, True)
            except TypeError:
                return True, repr(node)

        if node.is_expression_type():
            # Special handling if NPV and semi-NPV types:
            if not node.is_potentially_variable():
                return True, ftoa(node(), True)
            if node.__class__ is EXPR.MonomialTermExpression:
                return True, self._monomial_to_string(node)
            if node.__class__ is EXPR.LinearExpression:
                return True, self._linear_to_string(node)
            # we will descend into this, so type checking will happen later
            if node.is_component_type():
                self.treechecker(node)
            return False, None

        if node.is_component_type():
            if node.ctype not in valid_expr_ctypes_minlp:
                # Make sure all components in active constraints
                # are basic ctypes we know how to deal with.
                raise RuntimeError(
                    "Unallowable component '%s' of type %s found in an active "
                    "constraint or objective.\nThe GAMS writer cannot export "
                    "expressions with this component type."
                    % (node.name, node.ctype.__name__)
                )
            if node.ctype is not Var:
                # For these, make sure it's on the right model. We can check
                # Vars later since they don't disappear from the expressions
                self.treechecker(node)

        if node.is_fixed() and not (
            self.output_fixed_variables and node.is_potentially_variable()
        ):
            return True, ftoa(node(), True)
        else:
            assert node.is_variable_type()
            return True, self.smap.getSymbol(node)

    def _monomial_to_string(self, node):
        const, var = node.args
        if const.__class__ not in native_types:
            const = value(const)
        if var.is_fixed() and not self.output_fixed_variables:
            return ftoa(const * var.value, True)
        # Special handling: ftoa is slow, so bypass _to_string when this
        # is a trivial term
        if not const:
            return '0'
        if const in _plusMinusOne:
            if const < 0:
                return '-' + self.smap.getSymbol(var)
            else:
                return self.smap.getSymbol(var)
        return ftoa(const, True) + '*' + self.smap.getSymbol(var)

    def _linear_to_string(self, node):
        values = [
            (
                self._monomial_to_string(arg)
                if arg.__class__ is EXPR.MonomialTermExpression
                else (
                    ftoa(arg, True)
                    if arg.__class__ in native_numeric_types
                    else (
                        self.smap.getSymbol(arg)
                        if arg.is_variable_type()
                        and (not arg.fixed or self.output_fixed_variables)
                        else ftoa(value(arg), True)
                    )
                )
            )
            for arg in node.args
        ]
        return node._to_string(values, False, self.smap)


def expression_to_string(expr, treechecker, smap=None, output_fixed_variables=False):
    visitor = ToGamsVisitor(smap, treechecker, output_fixed_variables)
    expr_str = visitor.dfs_postorder_stack(expr)
    return expr_str, visitor.is_discontinuous


class Categorizer:
    """Class for representing categorized variables.

    Given a list of variable names and a symbol map, categorizes the variable
    names into the categories: binary, ints, positive and reals.

    """

    def __init__(self, var_list, symbol_map):
        self.binary = []
        self.ints = []
        self.positive = []
        self.reals = []
        self.fixed = []

        # categorize variables
        for var in var_list:
            v = symbol_map.getObject(var)
            if v.is_fixed():
                self.fixed.append(var)
            elif v.is_continuous():
                if v.lb == 0:
                    self.positive.append(var)
                else:
                    self.reals.append(var)
            elif v.is_integer():
                if all(bnd in _zero_one for bnd in v.bounds):
                    self.binary.append(var)
                else:
                    self.ints.append(var)
            else:
                raise RuntimeError(
                    "Cannot output variable to GAMS: effective variable "
                    "domain is not in {Reals, Integers, Binary}"
                )

    def __iter__(self):
        """Iterate over all variables.

        Yield a tuple containing the variables category and its name.
        """
        for category in ['binary', 'ints', 'positive', 'reals']:
            var_list = getattr(self, category)
            for var_name in var_list:
                yield category, var_name


class StorageTreeChecker:
    def __init__(self, model):
        # blocks are hashable so we can use a normal set
        self.tree = {model}
        self.model = model
        # add everything above the model
        pb = self.parent_block(model)
        while pb is not None:
            self.tree.add(pb)
            pb = self.parent_block(pb)

    def __call__(self, comp, exception_flag=True):
        if comp is self.model:
            return True

        # walk up tree until there are no more parents
        seen = set()
        pb = self.parent_block(comp)
        while pb is not None:
            if pb in self.tree:
                self.tree.update(seen)
                return True
            seen.add(pb)
            pb = self.parent_block(pb)

        if exception_flag:
            self.raise_error(comp)
        else:
            return False

    def parent_block(self, comp):
        if isinstance(comp, ICategorizedObject):
            parent = comp.parent
            while (parent is not None) and (not parent._is_heterogeneous_container):
                parent = parent.parent
            return parent
        else:
            return comp.parent_block()

    def raise_error(self, comp):
        raise RuntimeError(
            "GAMS writer: found component '%s' not on same model tree.\n"
            "All components must have the same parent model." % comp.name
        )


def split_long_line(line):
    """
    GAMS has an 80,000 character limit for lines, so split as many
    times as needed so as to not have illegal lines.
    """
    new_lines = ''
    while len(line) > 80000:
        i = 80000
        while line[i] != ' ':
            # Walk backwards to find closest space,
            # where it is safe to split to a new line
            if i < 0:
                raise RuntimeError("Found an 80,000+ character string with no spaces")
            i -= 1
        new_lines += line[:i] + '\n'
        # the space will be the first character in the next line,
        # so that the line doesn't start with the comment character '*'
        line = line[i:]
    new_lines += line
    return new_lines


class GAMSSymbolMap(SymbolMap):
    def __init__(self, var_labeler, var_list):
        super().__init__(self.var_label)
        self.var_labeler = var_labeler
        self.var_list = var_list

    def var_label(self, obj):
        # if obj.is_fixed():
        #    return str(value(obj))
        return self.getSymbol(obj, self.var_recorder)

    def var_recorder(self, obj):
        ans = self.var_labeler(obj)
        try:
            if obj.is_variable_type():
                self.var_list.append(ans)
        except:
            pass
        return ans


@WriterFactory.register('gams', 'Generate the corresponding GAMS file')
class ProblemWriter_gams(AbstractProblemWriter):
    def __init__(self):
        AbstractProblemWriter.__init__(self, ProblemFormat.gams)

    def __call__(self, model, output_filename, solver_capability, io_options):
        """
        Write a model in the GAMS modeling language format.

        Keyword Arguments
        -----------------
        output_filename: str
            Name of file to write GAMS model to. Optionally pass a file-like
            stream and the model will be written to that instead.
        io_options: dict
            - warmstart=True
                Warmstart by initializing model's variables to their values.
            - symbolic_solver_labels=False
                Use full Pyomo component names rather than
                shortened symbols (slower, but useful for debugging).
            - labeler=None
                Custom labeler. Incompatible with symbolic_solver_labels.
            - solver=None
                If None, GAMS will use default solver for model type.
            - mtype=None
                Model type. If None, will chose from lp, nlp, mip, and minlp.
            - add_options=None
                List of additional lines to write directly
                into model file before the solve statement.
                For model attributes, <model name> is GAMS_MODEL.
            - skip_trivial_constraints=False
                Skip writing constraints whose body section is fixed.
            - output_fixed_variables=False
                If True, output fixed variables as variables; otherwise,
                output numeric value.
            - file_determinism=1
                | How much effort do we want to put into ensuring the
                | GAMS file is written deterministically for a Pyomo model:
                |     0 : None
                |     1 : sort keys of indexed components (default)
                |     2 : sort keys AND sort names (over declaration order)
            - put_results=None
                Filename for optionally writing solution values and
                marginals.  If put_results_format is 'gdx', then GAMS
                will write solution values and marginals to
                GAMS_MODEL_p.gdx and solver statuses to
                {put_results}_s.gdx.  If put_results_format is 'dat',
                then solution values and marginals are written to
                (put_results).dat, and solver statuses to (put_results +
                'stat').dat.
            - put_results_format='gdx'
                Format used for put_results, one of 'gdx', 'dat'.

        """

        # Make sure not to modify the user's dictionary,
        # they may be reusing it outside of this call
        io_options = dict(io_options)

        # Use full Pyomo component names rather than
        # shortened symbols (slower, but useful for debugging).
        symbolic_solver_labels = io_options.pop("symbolic_solver_labels", False)

        # Custom labeler option. Incompatible with symbolic_solver_labels.
        labeler = io_options.pop("labeler", None)

        # If None, GAMS will use default solver for model type.
        solver = io_options.pop("solver", None)

        # If None, will chose from lp, nlp, mip, and minlp.
        mtype = io_options.pop("mtype", None)

        # Improved GAMS calling options
        solprint = io_options.pop("solprint", "off")
        limrow = io_options.pop("limrow", 0)
        limcol = io_options.pop("limcol", 0)
        solvelink = io_options.pop("solvelink", 5)

        # Lines to add before solve statement.
        add_options = io_options.pop("add_options", None)

        # Skip writing constraints whose body section is
        # fixed (i.e., no variables)
        skip_trivial_constraints = io_options.pop("skip_trivial_constraints", False)

        # Output fixed variables as variables
        output_fixed_variables = io_options.pop("output_fixed_variables", False)

        # How much effort do we want to put into ensuring the
        # GAMS file is written deterministically for a Pyomo model:
        #    0 : None
        #    1 : sort keys of indexed components (default)
        #    2 : sort keys AND sort names (over declaration order)
        file_determinism = io_options.pop("file_determinism", 1)
        sorter_map = {
            0: SortComponents.unsorted,
            1: SortComponents.deterministic,
            2: SortComponents.sortBoth,
        }
        sort = sorter_map[file_determinism]

        # Warmstart by initializing model's variables to their values.
        warmstart = io_options.pop("warmstart", True)

        # Filename for optionally writing solution values and marginals
        # Set to True by GAMSSolver
        put_results = io_options.pop("put_results", None)
        put_results_format = io_options.pop("put_results_format", 'gdx')
        assert put_results_format in ('gdx', 'dat')

        if len(io_options):
            raise ValueError(
                "GAMS writer passed unrecognized io_options:\n\t"
                + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
            )

        if solver is not None and solver.upper() not in valid_solvers:
            raise ValueError("GAMS writer passed unrecognized solver: %s" % solver)

        if mtype is not None:
            valid_mtypes = set(
                [
                    'lp',
                    'qcp',
                    'nlp',
                    'dnlp',
                    'rmip',
                    'mip',
                    'rmiqcp',
                    'rminlp',
                    'miqcp',
                    'minlp',
                    'rmpec',
                    'mpec',
                    'mcp',
                    'cns',
                    'emp',
                ]
            )
            if mtype.lower() not in valid_mtypes:
                raise ValueError(
                    "GAMS writer passed unrecognized model type: %s" % mtype
                )
            if (
                solver is not None
                and mtype.upper() not in valid_solvers[solver.upper()]
            ):
                raise ValueError(
                    "GAMS writer passed solver (%s) "
                    "unsuitable for given model type (%s)" % (solver, mtype)
                )

        if output_filename is None:
            output_filename = model.name + ".gms"

        if symbolic_solver_labels and (labeler is not None):
            raise ValueError(
                "GAMS writer: Using both the "
                "'symbolic_solver_labels' and 'labeler' "
                "I/O options is forbidden"
            )

        if symbolic_solver_labels:
            # Note that the Var and Constraint labelers must use the
            # same labeler, so that we can correctly detect name
            # collisions (which can arise when we truncate the labels to
            # the max allowable length.  GAMS requires all identifiers
            # to start with a letter.  We will (randomly) choose "s_"
            # (for 'shortened')
            var_labeler = con_labeler = ShortNameLabeler(
                60,
                prefix='s_',
                suffix='_',
                caseInsensitive=True,
                legalRegex='^[a-zA-Z]',
            )
        elif labeler is None:
            var_labeler = NumericLabeler('x')
            con_labeler = NumericLabeler('c')
        else:
            var_labeler = con_labeler = labeler

        var_list = []

        symbolMap = GAMSSymbolMap(var_labeler, var_list)

        # when sorting, there are a non-trivial number of
        # temporary objects created. these all yield
        # non-circular references, so disable GC - the
        # overhead is non-trivial, and because references
        # are non-circular, everything will be collected
        # immediately anyway.
        with PauseGC() as pgc:
            try:
                if isinstance(output_filename, str):
                    output_file = open(output_filename, "w")
                else:
                    # Support passing of stream such as a StringIO
                    # on which to write the model file
                    output_file = output_filename
                self._write_model(
                    model=model,
                    output_file=output_file,
                    solver_capability=solver_capability,
                    var_list=var_list,
                    var_label=symbolMap.var_label,
                    symbolMap=symbolMap,
                    con_labeler=con_labeler,
                    sort=sort,
                    skip_trivial_constraints=skip_trivial_constraints,
                    output_fixed_variables=output_fixed_variables,
                    warmstart=warmstart,
                    solver=solver,
                    mtype=mtype,
                    solprint=solprint,
                    limrow=limrow,
                    limcol=limcol,
                    solvelink=solvelink,
                    add_options=add_options,
                    put_results=put_results,
                    put_results_format=put_results_format,
                )
            finally:
                if isinstance(output_filename, str):
                    output_file.close()

        return output_filename, symbolMap

    def _write_model(
        self,
        model,
        output_file,
        solver_capability,
        var_list,
        var_label,
        symbolMap,
        con_labeler,
        sort,
        skip_trivial_constraints,
        output_fixed_variables,
        warmstart,
        solver,
        mtype,
        solprint,
        limrow,
        limcol,
        solvelink,
        add_options,
        put_results,
        put_results_format,
    ):
        constraint_names = []
        ConstraintIO = StringIO()
        linear = True
        linear_degree = set([0, 1])
        dnlp = False

        # Make sure there are no strange ActiveComponents. The expression
        # walker will handle strange things in constraints later.
        model_ctypes = model.collect_ctypes(active=True)
        invalids = set()
        for t in model_ctypes - valid_active_ctypes_minlp:
            if issubclass(t, ActiveComponent):
                invalids.add(t)
        if len(invalids):
            invalids = [t.__name__ for t in invalids]
            raise RuntimeError(
                "Unallowable active component(s) %s.\nThe GAMS writer cannot "
                "export models with this component type." % ", ".join(invalids)
            )

        tc = StorageTreeChecker(model)

        # Walk through the model and generate the constraint definition
        # for all active constraints.  Any Vars / Expressions that are
        # encountered will be added to the var_list due to the labeler
        # defined above.
        for con in model.component_data_objects(Constraint, active=True, sort=sort):
            lb, body, ub = con.to_bounded_expression(True)
            if lb is None and ub is None:
                assert not con.equality
                continue  # non-binding, so skip

            con_body = as_numeric(body)
            if skip_trivial_constraints and con_body.is_fixed():
                continue
            if linear:
                if con_body.polynomial_degree() not in linear_degree:
                    linear = False

            cName = symbolMap.getSymbol(con, con_labeler)
            con_body_str, con_discontinuous = expression_to_string(
                con_body,
                tc,
                smap=symbolMap,
                output_fixed_variables=output_fixed_variables,
            )
            dnlp |= con_discontinuous
            if con.equality:
                constraint_names.append('%s' % cName)
                ConstraintIO.write(
                    '%s.. %s =e= %s ;\n'
                    % (constraint_names[-1], con_body_str, ftoa(ub, False))
                )
            else:
                if lb is not None:
                    constraint_names.append('%s_lo' % cName)
                    ConstraintIO.write(
                        '%s.. %s =l= %s ;\n'
                        % (constraint_names[-1], ftoa(lb, False), con_body_str)
                    )
                if ub is not None:
                    constraint_names.append('%s_hi' % cName)
                    ConstraintIO.write(
                        '%s.. %s =l= %s ;\n'
                        % (constraint_names[-1], con_body_str, ftoa(ub, False))
                    )

        obj = list(model.component_data_objects(Objective, active=True, sort=sort))
        if len(obj) != 1:
            raise RuntimeError(
                "GAMS writer requires exactly one active objective (found %s)"
                % (len(obj))
            )
        obj = obj[0]
        if linear:
            if obj.polynomial_degree() not in linear_degree:
                linear = False
        obj_expr_str, obj_discontinuous = expression_to_string(
            obj.expr, tc, smap=symbolMap, output_fixed_variables=output_fixed_variables
        )
        dnlp |= obj_discontinuous
        oName = symbolMap.getSymbol(obj, con_labeler)
        constraint_names.append(oName)
        ConstraintIO.write('%s.. GAMS_OBJECTIVE =e= %s ;\n' % (oName, obj_expr_str))

        # Categorize the variables that we found
        categorized_vars = Categorizer(var_list, symbolMap)

        # Write the GAMS model
        output_file.write("$offlisting\n")
        # $offdigit ignores extra precise digits instead of erroring
        output_file.write("$offdigit\n\n")
        output_file.write("EQUATIONS\n\t")
        output_file.write("\n\t".join(constraint_names))
        if categorized_vars.binary:
            output_file.write(";\n\nBINARY VARIABLES\n\t")
            output_file.write("\n\t".join(categorized_vars.binary))
        if categorized_vars.ints:
            output_file.write(";\n\nINTEGER VARIABLES")
            output_file.write("\n\t")
            output_file.write("\n\t".join(categorized_vars.ints))
        if categorized_vars.positive:
            output_file.write(";\n\nPOSITIVE VARIABLES\n\t")
            output_file.write("\n\t".join(categorized_vars.positive))
        output_file.write(";\n\nVARIABLES\n\tGAMS_OBJECTIVE\n\t")
        output_file.write("\n\t".join(categorized_vars.reals + categorized_vars.fixed))
        output_file.write(";\n\n")

        for var in categorized_vars.fixed:
            output_file.write(
                "%s.fx = %s;\n" % (var, ftoa(value(symbolMap.getObject(var)), False))
            )
        output_file.write("\n")

        for line in ConstraintIO.getvalue().splitlines():
            if len(line) > 80000:
                line = split_long_line(line)
            output_file.write(line + "\n")

        output_file.write("\n")

        warn_int_bounds = False
        for category, var_name in categorized_vars:
            var = symbolMap.getObject(var_name)
            tc(var)
            lb, ub = var.bounds
            if category == 'positive':
                if ub is not None:
                    output_file.write("%s.up = %s;\n" % (var_name, ftoa(ub, False)))
            elif category == 'ints':
                if lb is None:
                    warn_int_bounds = True
                    # GAMS doesn't allow -INF lower bound for ints
                    logger.warning(
                        "Lower bound for integer variable %s set "
                        "to -1.0E+100." % var.name
                    )
                    output_file.write("%s.lo = -1.0E+100;\n" % (var_name))
                elif lb != 0:
                    output_file.write("%s.lo = %s;\n" % (var_name, ftoa(lb, False)))
                if ub is None:
                    warn_int_bounds = True
                    # GAMS has an option value called IntVarUp that is the
                    # default upper integer bound, which it applies if the
                    # integer's upper bound is INF. This option maxes out at
                    # 2147483647, so we can go higher by setting the bound.
                    logger.warning(
                        "Upper bound for integer variable %s set "
                        "to +1.0E+100." % var.name
                    )
                    output_file.write("%s.up = +1.0E+100;\n" % (var_name))
                else:
                    output_file.write("%s.up = %s;\n" % (var_name, ftoa(ub, False)))
            elif category == 'binary':
                if lb != 0:
                    output_file.write("%s.lo = %s;\n" % (var_name, ftoa(lb, False)))
                if ub != 1:
                    output_file.write("%s.up = %s;\n" % (var_name, ftoa(ub, False)))
            elif category == 'reals':
                if lb is not None:
                    output_file.write("%s.lo = %s;\n" % (var_name, ftoa(lb, False)))
                if ub is not None:
                    output_file.write("%s.up = %s;\n" % (var_name, ftoa(ub, False)))
            else:
                raise KeyError('Category %s not supported' % category)
            if warmstart and var.value is not None:
                output_file.write("%s.l = %s;\n" % (var_name, ftoa(var.value, False)))

        if warn_int_bounds:
            logger.warning(
                "GAMS requires finite bounds for integer variables. 1.0E100 "
                "is as extreme as GAMS will define, and should be enough to "
                "appear unbounded. If the solver cannot handle this bound, "
                "explicitly set a smaller bound on the pyomo model, or try a "
                "different GAMS solver."
            )

        model_name = "GAMS_MODEL"
        output_file.write("\nMODEL %s /all/ ;\n" % model_name)

        if mtype is None:
            mtype = ('lp', 'nlp', 'mip', 'minlp')[
                (0 if linear else 1)
                + (2 if (categorized_vars.binary or categorized_vars.ints) else 0)
            ]
            if mtype == 'nlp' and dnlp:
                mtype = 'dnlp'

        if solver is not None:
            if mtype.upper() not in valid_solvers[solver.upper()]:
                raise ValueError(
                    "GAMS writer passed solver (%s) "
                    "unsuitable for model type (%s)" % (solver, mtype)
                )
            output_file.write("option %s=%s;\n" % (mtype, solver))

        output_file.write("option solprint=%s;\n" % solprint)
        output_file.write("option limrow=%d;\n" % limrow)
        output_file.write("option limcol=%d;\n" % limcol)
        output_file.write("option solvelink=%d;\n" % solvelink)

        if put_results is not None and put_results_format == 'gdx':
            output_file.write("option savepoint=1;\n")

        if add_options is not None:
            output_file.write("\n* START USER ADDITIONAL OPTIONS\n")
            for line in add_options:
                output_file.write('\n' + line)
            output_file.write("\n\n* END USER ADDITIONAL OPTIONS\n\n")

        output_file.write(
            "SOLVE %s USING %s %simizing GAMS_OBJECTIVE;\n\n"
            % (model_name, mtype, 'min' if obj.sense == minimize else 'max')
        )

        # Set variables to store certain statuses and attributes
        stat_vars = [
            'MODELSTAT',
            'SOLVESTAT',
            'OBJEST',
            'OBJVAL',
            'NUMVAR',
            'NUMEQU',
            'NUMDVAR',
            'NUMNZ',
            'ETSOLVE',
        ]
        output_file.write(
            "Scalars MODELSTAT 'model status', SOLVESTAT 'solve status';\n"
        )
        output_file.write("MODELSTAT = %s.modelstat;\n" % model_name)
        output_file.write("SOLVESTAT = %s.solvestat;\n\n" % model_name)

        output_file.write("Scalar OBJEST 'best objective', OBJVAL 'objective value';\n")
        output_file.write("OBJEST = %s.objest;\n" % model_name)
        output_file.write("OBJVAL = %s.objval;\n\n" % model_name)

        output_file.write("Scalar NUMVAR 'number of variables';\n")
        output_file.write("NUMVAR = %s.numvar\n\n" % model_name)

        output_file.write("Scalar NUMEQU 'number of equations';\n")
        output_file.write("NUMEQU = %s.numequ\n\n" % model_name)

        output_file.write("Scalar NUMDVAR 'number of discrete variables';\n")
        output_file.write("NUMDVAR = %s.numdvar\n\n" % model_name)

        output_file.write("Scalar NUMNZ 'number of nonzeros';\n")
        output_file.write("NUMNZ = %s.numnz\n\n" % model_name)

        output_file.write("Scalar ETSOLVE 'time to execute solve statement';\n")
        output_file.write("ETSOLVE = %s.etsolve\n\n" % model_name)

        if put_results is not None:
            if put_results_format == 'gdx':
                output_file.write("\nexecute_unload '%s_s.gdx'" % put_results)
                for stat in stat_vars:
                    output_file.write(", %s" % stat)
                output_file.write(";\n")
            else:
                results = put_results + '.dat'
                output_file.write("\nfile results /'%s'/;" % results)
                output_file.write("\nresults.nd=15;")
                output_file.write("\nresults.nw=21;")
                output_file.write("\nput results;")
                output_file.write("\nput 'SYMBOL  :  LEVEL  :  MARGINAL' /;")
                for var in var_list:
                    output_file.write("\nput %s ' ' %s.l ' ' %s.m /;" % (var, var, var))
                for con in constraint_names:
                    output_file.write("\nput %s ' ' %s.l ' ' %s.m /;" % (con, con, con))
                output_file.write(
                    "\nput GAMS_OBJECTIVE ' ' GAMS_OBJECTIVE.l "
                    "' ' GAMS_OBJECTIVE.m;\n"
                )

                statresults = put_results + 'stat.dat'
                output_file.write("\nfile statresults /'%s'/;" % statresults)
                output_file.write("\nstatresults.nd=15;")
                output_file.write("\nstatresults.nw=21;")
                output_file.write("\nput statresults;")
                output_file.write("\nput 'SYMBOL   :   VALUE' /;")
                for stat in stat_vars:
                    output_file.write("\nput '%s' ' ' %s /;\n" % (stat, stat))


valid_solvers = {
    'ALPHAECP': {'MINLP', 'MIQCP'},
    'AMPL': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
    },
    'ANTIGONE': {'NLP', 'CNS', 'DNLP', 'RMINLP', 'MINLP', 'QCP', 'MIQCP', 'RMIQCP'},
    'BARON': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'BDMLP': {'LP', 'MIP', 'RMIP'},
    'BDMLPD': {'LP', 'RMIP'},
    'BENCH': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'BONMIN': {'MINLP', 'MIQCP'},
    'BONMINH': {'MINLP', 'MIQCP'},
    'CBC': {'LP', 'MIP', 'RMIP'},
    'COINBONMIN': {'MINLP', 'MIQCP'},
    'COINCBC': {'LP', 'MIP', 'RMIP'},
    'COINCOUENNE': {'NLP', 'CNS', 'DNLP', 'RMINLP', 'MINLP', 'QCP', 'MIQCP', 'RMIQCP'},
    'COINIPOPT': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'COINOS': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'COINSCIP': {
        'MIP',
        'NLP',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'CONOPT': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'CONOPT3': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'CONOPT4': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'CONOPTD': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'COPT': {'LP', 'MIP', 'RMIP', 'QCP', 'MIQCP', 'RMIQCP'},
    'CONVERT': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'CONVERTD': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
        'EMP',
    },
    'COUENNE': {'NLP', 'CNS', 'DNLP', 'RMINLP', 'MINLP', 'QCP', 'MIQCP', 'RMIQCP'},
    'CPLEX': {'LP', 'MIP', 'RMIP', 'QCP', 'MIQCP', 'RMIQCP'},
    'CPLEXD': {'LP', 'MIP', 'RMIP', 'QCP', 'MIQCP', 'RMIQCP'},
    'CPOPTIMIZER': {'MIP', 'MINLP', 'MIQCP'},
    'DE': {'EMP'},
    'DECIS': {'EMP'},
    'DECISC': {'LP'},
    'DECISM': {'LP'},
    'DICOPT': {'MINLP', 'MIQCP'},
    'DICOPTD': {'MINLP', 'MIQCP'},
    'EXAMINER': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'EXAMINER2': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'GAMSCHK': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'GLOMIQO': {'QCP', 'MIQCP', 'RMIQCP'},
    'GUROBI': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'DNLP',
        'MINLP',
        'RMINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'GUSS': {'LP', 'MIP', 'NLP', 'MCP', 'CNS', 'DNLP', 'MINLP', 'QCP', 'MIQCP'},
    'IPOPT': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'IPOPTH': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'JAMS': {'EMP'},
    'HIGHS': {'LP', 'MIP', 'RMIP'},
    'KESTREL': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
        'EMP',
    },
    'KNITRO': {
        'LP',
        'RMIP',
        'NLP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'LGO': {'LP', 'RMIP', 'NLP', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'LGOD': {'LP', 'RMIP', 'NLP', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'LINDO': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
        'EMP',
    },
    'LINDOGLOBAL': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'LINGO': {'LP', 'MIP', 'RMIP', 'NLP', 'DNLP', 'RMINLP', 'MINLP'},
    'LOCALSOLVER': {
        'MIP',
        'NLP',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'LOGMIP': {'EMP'},
    'LS': {'LP', 'RMIP'},
    'MILES': {'MCP'},
    'MILESE': {'MCP'},
    'MINOS': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'MINOS5': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'MINOS55': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'MOSEK': {'LP', 'MIP', 'RMIP', 'NLP', 'DNLP', 'RMINLP', 'QCP', 'MIQCP', 'RMIQCP'},
    'MPECDUMP': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
    },
    'MPSGE': {},
    'MSNLP': {'NLP', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'NLPEC': {'MCP', 'MPEC', 'RMPEC'},
    'ODHCPLEX': {'MINLP'},
    'OQNLP': {'NLP', 'DNLP', 'MINLP', 'QCP', 'MIQCP'},
    'OS': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'OSICPLEX': {'LP', 'MIP', 'RMIP'},
    'OSIGUROBI': {'LP', 'MIP', 'RMIP'},
    'OSIMOSEK': {'LP', 'MIP', 'RMIP'},
    'OSISOPLEX': {'LP', 'RMIP'},
    'OSIXPRESS': {'LP', 'MIP', 'RMIP'},
    'PATH': {'MCP', 'CNS'},
    'PATHC': {'MCP', 'CNS'},
    'PATHNLP': {'LP', 'RMIP', 'NLP', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'PYOMO': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'MPEC',
        'RMPEC',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
    },
    'QUADMINOS': {'LP'},
    'SBB': {'MINLP', 'MIQCP'},
    'SCENSOLVER': {
        'LP',
        'MIP',
        'RMIP',
        'NLP',
        'MCP',
        'CNS',
        'DNLP',
        'RMINLP',
        'MINLP',
        'QCP',
        'MIQCP',
        'RMIQCP',
    },
    'SCIP': {'MIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'MINLP', 'QCP', 'MIQCP', 'RMIQCP'},
    'SHOT': {'MINLP', 'MIQCP'},
    'SNOPT': {'LP', 'RMIP', 'NLP', 'CNS', 'DNLP', 'RMINLP', 'QCP', 'RMIQCP'},
    'SOPLEX': {'LP', 'RMIP'},
    'XA': {'LP', 'MIP', 'RMIP'},
    'XPRESS': {'LP', 'MIP', 'RMIP', 'QCP', 'MIQCP', 'RMIQCP'},
}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""This module provides general utilities for producing formatted I/O

.. autosummary::

   tostr
   tabular_writer
   wrap_reStructuredText
   StreamIndenter
"""

import io
import re
import types

from typing import Iterable

from pyomo.common.sorting import sorted_robust


def tostr(value, quote_str=False):
    """Convert a value to a string

    This function is a thin wrapper around `str(value)` to resolve a
    problematic __str__ implementation in the standard Python container
    types (tuple, list, and dict).  Those classes implement __str__ the
    same as __repr__ (by calling repr() on each contained object).  That
    is frequently undesirable, as you may wish the string representation
    of a container to contain the string representations of the
    contained objects.

    This function generates string representations for native Python
    containers (tuple, list, and dict) that contains the string
    representations of the contained objects.  In addition, it also
    applies the same special handling to any types that derive from the
    standard containers without overriding either __repn__ or __str__.

    Parameters
    ----------
    value: object
        the object to convert to a string
    quote_str: bool
        if True, and if `value` is a `str`, then return a "quoted
        string" (as generated by repr()).  This is primarily used when
        recursively processing native Python containers.

    Returns
    -------
    str

    """
    # Override the generation of str(list), but only if the object is
    # using the default implementation of list.__str__.  Note that the
    # default implementation of __str__ (in CPython) is to call __repr__,
    # so we will test both.  This is particularly important for
    # collections.namedtuple, which reimplements __repr__ but not
    # __str__.
    _type = type(value)
    if _type not in tostr.handlers:
        # Default to the None handler (just call str()), but override it
        # in particular instances:
        tostr.handlers[_type] = tostr.handlers[None]
        if isinstance(value, list):
            if _type.__str__ is list.__str__ and _type.__repr__ is list.__repr__:
                tostr.handlers[_type] = tostr.handlers[list]
        elif isinstance(value, tuple):
            if _type.__str__ is tuple.__str__ and _type.__repr__ is tuple.__repr__:
                tostr.handlers[_type] = tostr.handlers[tuple]
        elif isinstance(value, dict):
            if _type.__str__ is dict.__str__ and _type.__repr__ is dict.__repr__:
                tostr.handlers[_type] = tostr.handlers[dict]
        elif isinstance(value, str):
            tostr.handlers[_type] = tostr.handlers[str]

    return tostr.handlers[_type](value, quote_str)


tostr.handlers = {
    list: lambda value, quote_str: (
        "[%s]" % (', '.join(tostr(v, True) for v in value))
    ),
    dict: lambda value, quote_str: (
        "{%s}"
        % (
            ', '.join(
                '%s: %s' % (tostr(k, True), tostr(v, True)) for k, v in value.items()
            )
        )
    ),
    tuple: lambda value, quote_str: (
        "(%s,)" % (tostr(value[0], True),)
        if len(value) == 1
        else "(%s)" % (', '.join(tostr(v, True) for v in value))
    ),
    str: lambda value, quote_str: (repr(value) if quote_str else value),
    None: lambda value, quote_str: str(value),
}


def tabular_writer(ostream, prefix, data, header, row_generator):
    """Output data in tabular form

    Parameters
    ----------
    ostream: io.TextIOBase
        the stream to write to
    prefix: str
        prefix each generated line with this string
    data: iterable
        an iterable object that returns (key, value) pairs
        (e.g., from iteritems()) defining each row in the table
    header: List[str]
        list of column headers
    row_generator: function
        a function that accepts the `key` and `value` from `data` and
        returns either a tuple defining the entries for a single row, or
        a generator that returns a sequence of table rows to be output
        for the specified `key`

    """

    prefix = tostr(prefix)

    _rows = {}
    # NB: _width is a list because we will change these values
    if header:
        header = (u"Key",) + tuple(tostr(x) for x in header)
        _width = [len(x) for x in header]
    else:
        _width = None
    _minWidth = 0

    for _key, _val in data:
        try:
            _rowSet = row_generator(_key, _val)
            if isinstance(_rowSet, types.GeneratorType):
                _rowSet = list(_rowSet)
            else:
                _rowSet = [_rowSet]
        except ValueError:
            # A ValueError can be raised when row_generator is called
            # (if it is a function), or when it is exhausted generating
            # the list (if it is a generator)
            _minWidth = 4  # Ensure columns are wide enough to output "None"
            _rows[_key] = None
            continue

        # Include the key for only the first line in a rowset, and only
        # if we printed out a header (if there is no header, then the
        # key is not included)
        _rows[_key] = [
            (("" if i else tostr(_key),) if header else ())
            + tuple(tostr(x) for x in _r)
            for i, _r in enumerate(_rowSet)
        ]

        if not _rows[_key]:
            _minWidth = 4
        elif not _width:
            _width = [0] * len(_rows[_key][0])
        for _row in _rows[_key]:
            for col, x in enumerate(_row):
                _width[col] = max(_width[col], len(x), col and _minWidth)

    # NB: left-justify header entries
    if header:
        # Note: do not right-pad the last header with unnecessary spaces
        tmp = _width[-1]
        _width[-1] = 0
        ostream.write(
            prefix
            + " : ".join("%%-%ds" % _width[i] % x for i, x in enumerate(header))
            + "\n"
        )
        _width[-1] = tmp

    # If there is no data, we are done...
    if not _rows:
        return

    # right-justify data, except for the last column if there are spaces
    # in the data (probably an expression or vector)
    _width = ["%" + str(i) + "s" for i in _width]

    if any(' ' in r[-1] for x in _rows.values() if x is not None for r in x):
        _width[-1] = '%s'
    for _key in _rows:
        _rowSet = _rows[_key]
        if not _rowSet:
            _rowSet = [[_key] + [None] * (len(_width) - 1)]
        for _data in _rowSet:
            ostream.write(
                prefix + " : ".join(_width[i] % x for i, x in enumerate(_data)) + "\n"
            )


class StreamIndenter:
    """
    Mock-up of a file-like object that wraps another file-like object
    and indents all data using the specified string before passing it to
    the underlying file.  Since this presents a full file interface,
    StreamIndenter objects may be arbitrarily nested.
    """

    _newline_re = re.compile('\n([^\n])')
    _blankline_re = re.compile('\n\n')

    def __init__(self, ostream: io.TextIOBase, indent: str = ' ' * 4):
        super().__setattr__('wrapped_os', ostream)
        # The following is a "cute" trick: because of the __getattr__ /
        # __setattr__ overloads, nested StreamIndenter instances all
        # print directly to the underlying stream object, and all share
        # a common `newline` flag.
        if isinstance(ostream, StreamIndenter):
            super().__setattr__('target_os', ostream.target_os)
            indent = ostream.indent + indent
        else:
            super().__setattr__('target_os', ostream)
            # We will assume the last thing written to the stream we
            # are wrapping ended with a newline
            super().__setattr__('newline', True)
        super().__setattr__('indent', indent)
        super().__setattr__('indent_match', f'\n{indent}\\1')
        super().__setattr__('stripped_indent', indent.rstrip())
        if self.stripped_indent:
            super().__setattr__('blankline_match', f'\n{self.stripped_indent}\n')

    def __getattr__(self, name: str):
        return getattr(self.wrapped_os, name)

    def __setattr__(self, name: str, val):
        if name in self.__dict__:
            super().__setattr__(name, val)
        else:
            self.wrapped_os.__setattr__(name, val)

    def write(self, data: str) -> int:
        if not data:
            return 0
        written = 0
        if self.newline:
            if data[0] != '\n':
                written += self.target_os.write(self.indent)
            elif self.stripped_indent:
                written += self.target_os.write(self.stripped_indent)
        data = self._newline_re.sub(self.indent_match, data)
        if self.stripped_indent:
            data, n = self._blankline_re.subn(self.blankline_match, data)
            # If we replaced any blank lines, then we need to check
            # again to catch cases like "\n\n\n"
            if n:
                data = self._blankline_re.sub(self.blankline_match, data)
        written += self.target_os.write(data)
        self.newline = data.endswith('\n')
        return written

    def writelines(self, sequence: Iterable[str]) -> None:
        for x in sequence:
            self.write(x)


_indentation_re = re.compile(r'\s*')
_bullet_re = re.compile(
    r'([-+*] +)'  # bulleted lists
    r'|(\(?[0-9]+[\)\.] +)'  # enumerated lists (arabic numerals)
    r'|(\(?[ivxlcdm]+[\)\.] +)'  # enumerated lists (roman numerals)
    r'|(\(?[IVXLCDM]+[\)\.] +)'  # enumerated lists (roman numerals)
    r'|(\(?[a-zA-Z][\)\.] +)'  # enumerated lists (letters)
    r'|(\(?\#[\)\.] +)'  # auto enumerated lists
    r'|([a-zA-Z0-9_ ]+ +: +)'  # definitions
    r'|(:[a-zA-Z0-9_ ]+: +)'  # field name
    r'|(?:\[\s*[A-Za-z0-9\.]+\s*\] +)'  # [PASS]|[FAIL]|[ OK ]
)
_verbatim_line_start = re.compile(
    r'(\| )'  # line blocks
    r'|(\+((-{3,})|(={3,}))\+)'  # grid table
)
_verbatim_line = re.compile(
    r'(={3,}[ =]+)'  # simple tables, ======== sections
    # sections
    + ''.join(r'|(\%s{3,})' % c for c in r'!"#$%&\'()*+,-./:;<>?@[\\]^_`{|}~')
)


def wrap_reStructuredText(docstr, wrapper):
    """A text wrapper that honors paragraphs and basic reStructuredText markup

    This wraps `textwrap.fill()` to first separate the incoming text by
    paragraphs before using ``wrapper`` to wrap each one.  It includes a
    basic (partial) parser for reStructuredText format to attempt to
    avoid wrapping structural elements like section headings, bullet /
    enumerated lists, and tables.

    Parameters
    ----------
    docstr : str
        The incoming string to parse and wrap

    wrapper : `textwrap.TextWrap`
        The configured `TextWrap` object to use for wrapping paragraphs.
        While the object will be reconfigured within this function, it
        will be restored to its original state upon exit.

    """
    # As textwrap only works on single paragraphs, we need to break
    # up the incoming message into paragraphs before we pass it to
    # textwrap.
    paragraphs = [(None, None, None)]
    literal_block = False
    verbatim = False
    for line in docstr.rstrip().splitlines():
        leading = _indentation_re.match(line).group()
        content = line.strip()
        if not content:
            if literal_block:
                if literal_block[0] == 2:
                    literal_block = False
            elif paragraphs[-1][2] and ''.join(paragraphs[-1][2]).endswith('::'):
                literal_block = (0, paragraphs[-1][1])
            paragraphs.append((None, None, None))
            continue
        if literal_block:
            if literal_block[0] == 0:
                if len(literal_block[1]) < len(leading):
                    # indented literal block
                    literal_block = 1, leading
                    paragraphs.append((None, None, line))
                    continue
                elif (
                    len(literal_block[1]) == len(leading)
                    and content[0] in '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
                ):
                    # quoted literal block
                    literal_block = 2, leading
                    paragraphs.append((None, None, line))
                    continue
                else:
                    # invalid literal block
                    literal_block = False
            elif leading.startswith(literal_block[1]):
                paragraphs.append((None, None, line))
                continue
            else:
                # fall back on normal line processing
                literal_block = False
        if content == '```':
            # Not part of ReST, but we have supported this in Pyomo for a long time
            verbatim ^= True
        elif verbatim:
            paragraphs.append((None, None, line))
        elif _verbatim_line_start.match(content):
            # This catches lines that start with patterns that indicate
            # that the line should not be wrapped (line blocks, grid
            # tables)
            paragraphs.append((None, None, line))
        elif _verbatim_line.match(content):
            # This catches whole line patterns that should not be
            # wrapped with previous/subsequent lines (e.g., simple table
            # headers, section headers)
            paragraphs.append((None, None, line))
        else:
            matchBullet = _bullet_re.match(content)
            if matchBullet:
                # Handle things that look like bullet lists specially
                hang = matchBullet.group()
                paragraphs.append((leading, leading + ' ' * len(hang), [content]))
            elif paragraphs[-1][1] == leading:
                # Continuing a text block
                paragraphs[-1][2].append(content)
            else:
                # Beginning a new text block
                paragraphs.append((leading, leading, [content]))

    while paragraphs and paragraphs[0][2] is None:
        paragraphs.pop(0)

    wrapper_init = wrapper.initial_indent, wrapper.subsequent_indent
    try:
        for i, (indent, subseq, par) in enumerate(paragraphs):
            base_indent = wrapper_init[1] if i else wrapper_init[0]

            if indent is None:
                if par is None:
                    paragraphs[i] = ''
                else:
                    paragraphs[i] = base_indent + par
                continue

            wrapper.initial_indent = base_indent + indent
            wrapper.subsequent_indent = base_indent + subseq
            paragraphs[i] = wrapper.fill(' '.join(par))
    finally:
        # Avoid side-effects and restore the initial wrapper state
        wrapper.initial_indent, wrapper.subsequent_indent = wrapper_init

    return '\n'.join(paragraphs)
